import os
import unittest

from pdfconduit.convert import PDF2IMG
from tests import *


class TestPdf2Img(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pdf_path = os.path.join(test_data_dir, "plan_p.pdf")
        cls.img = None

    def tearDown(self):
        if os.path.exists(self.img):
            os.remove(self.img)

    def test_pdf2img(self):
        """Convert a PDF file to a png image."""
        img = PDF2IMG(self.pdf_path).save()

        # Assert img file exists
        self.assertTrue(os.path.exists(img[0]))

        # Assert img file is correct file type
        self.assertTrue(img[0].endswith(".png"))
        self.img = img[0]


if __name__ == "__main__":
    unittest.main()



from pdfconduit.utils.driver import PdfDriver
from pdfconduit.utils.info import Info
from pdfconduit.utils.path import add_suffix
from pdfconduit.utils.read import pypdf_reader

__all__ = [
    "add_suffix",
    "Info",
    "pypdf_reader",
    "PdfDriver",
]



import os
from tempfile import TemporaryDirectory, NamedTemporaryFile
from warnings import warn
from abc import ABC
from datetime import datetime
from io import BytesIO

from pypdf import PdfWriter, PdfReader

from pdfconduit.internals.exceptions import OutputException
from pdfconduit.utils import pypdf_reader, add_suffix
from pdfconduit.utils.typing import Optional, Any, Dict, Self, Union


class BaseConduit(ABC):
    _path: Optional[str] = None
    _metadata: Dict[str, Any] = {}

    output: Optional[str] = None
    _output_dir: Optional[str] = None
    _closed: bool = False

    _pdf_file = None
    _stream: BytesIO = None
    _reader: PdfReader
    _writer: PdfWriter

    _tempdir: Optional[TemporaryDirectory] = None
    _tempfile: Optional[NamedTemporaryFile] = None

    def __init__(
        self, pdf: Union[str, BytesIO], decrypt_pw: Optional[str] = None
    ) -> None:
        self._decrypt_pw = decrypt_pw

        if isinstance(pdf, BytesIO):
            self._read_stream(pdf)
        else:
            # Open the file & instantiate a PDF reader
            self._path = pdf
            self._open_and_read()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            raise exc_val
        self.write()
        return True

    def _open_and_read(self) -> Self:
        self._pdf_file = open(self._path, "rb")
        self._reader: PdfReader = pypdf_reader(self._pdf_file, self._decrypt_pw)
        self._writer: PdfWriter = PdfWriter(clone_from=self._reader)
        return self

    def _read_stream(self, stream: BytesIO) -> Self:
        self._stream = stream
        self._reader = PdfReader(self._stream)
        self._writer: PdfWriter = PdfWriter(clone_from=self._reader)
        return self

    def write(self):
        # Set default output in case none was set
        self._set_default_output("modified")

        # Add metadata
        # Format the current date and time for the metadata
        utc_time = "-05'00'"  # UTC time optional
        time = datetime.now().strftime(f"D\072%Y%m%d%H%M%S{utc_time}")
        default_metadata = {
            "/Producer": "pdfconduit",
            "/Creator": "pdfconduit",
            "/Author": "pdfconduit",
            "/ModDate": time,
        }
        default_metadata.update(self._metadata)
        self._writer.add_metadata(default_metadata)

        # Close the PDF reader & file reader
        self._reader.close()
        if self._pdf_file is not None:
            self._pdf_file.close()
        if self._stream is not None:
            self._stream.close()

        # Confirm output path is set
        if self.output is None:
            raise OutputException

        # Write the PDF to the output file
        if self._tempdir is not None:
            self._writer.write(self.output)
            self._tempfile.close()
        else:
            with open(self.output, "wb") as output_pdf:
                self._writer.write(output_pdf)

        self._writer.close()

        self._closed = True

        return self.output

    def cleanup(self) -> Self:
        if self._tempdir is not None:
            self._tempdir.cleanup()
        return self

    def write_to_stream(self):
        # todo: implement
        pass

    def set_metadata(self, metadata: Dict[str, Any]) -> Self:
        self._metadata = metadata
        return self

    def set_output(self, output: str) -> Self:
        self.output = output
        return self

    def set_output_suffix(self, suffix: str) -> Self:
        return self.set_output(
            add_suffix(
                (
                    os.path.join(self._output_dir, os.path.basename(self._path))
                    if self._output_dir is not None
                    else self._path
                ),
                suffix,
            )
        )

    def set_output_directory(self, directory: str) -> Self:
        self._output_dir = directory
        return self

    def set_output_temp(
        self, tempdir: Optional[TemporaryDirectory] = None, suffix: str = ""
    ) -> Self:
        self._tempdir = tempdir if tempdir else TemporaryDirectory(prefix="pdfconduit_")
        self._tempfile = NamedTemporaryFile(
            suffix="_" + suffix.replace("_", "") + ".pdf",
            dir=self._tempdir.name,
            delete=False,
        )
        return self.set_output(self._tempfile.name)

    def _set_default_output(self, suffix: str) -> None:
        if self.output is None:
            if self._path is None:
                warn(
                    """
                    Saving PDFs to a temporary directory because an original file directory cannot be determined
                    (likely because we're reading from a stream).  Add a call to `conduit.cleanup() after PDF
                    processing to delete the created temporary directory.
                    """,
                    UserWarning,
                )
                self.set_output_temp(suffix=suffix)
                return
            self.set_output_suffix(suffix)



from typing import Optional

from pdfconduit.watermark.modify import LETTER


class CanvasStr:
    """Canvas string data object used for storing canvas.drawString parameters."""

    def __init__(
        self,
        string: str,
        font: str = "Vera",
        color: str = "black",
        size: int = 40,
        opacity: float = 0.1,
        x: float = 0,
        y: float = 0,
        x_centered: bool = True,
        y_centered: bool = False,
    ):
        self.string = string
        self.font = font
        self.color = color
        self.size = size
        self.opacity = opacity
        self.x = x
        self.y = y
        self.x_centered = x_centered
        self.y_centered = y_centered


class CanvasImg:
    """Canvas image data object used for storing canvas.drawImage parameters."""

    def __init__(
        self,
        image: str,
        opacity: float = 0.1,
        x: float = 0,
        y: float = 0,
        w: int = LETTER[0],
        h: int = LETTER[1],
        mask: Optional[str] = "auto",
        preserve_aspect_ratio: bool = True,
        centered: bool = False,
    ):
        self.image = image
        self.opacity = opacity
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.mask = mask
        self.preserve_aspect_ratio = preserve_aspect_ratio
        self.centered = centered


class CanvasObjects:
    """Canvas object collector to store list of canvas objects."""

    def __init__(self):
        self.objects = []

    def __iter__(self) -> iter:
        return iter(self.objects)

    def add(self, canvas_object) -> None:
        self.objects.append(canvas_object)



from pypdf import PdfReader


def pypdf_reader(pdf, decrypt=None) -> PdfReader:
    """
    Retrieve a PdfFileReader object that has been decrypted if a password is specified.

    :param pdf: PDF document to read
    :param decrypt: Owner password to decrypt pdf
    :return: PdfFileReader object
    """
    if decrypt:
        reader = PdfReader(pdf)
        reader.decrypt(decrypt)
        return reader
    else:
        return PdfReader(pdf)



import inspect
import json
import os
import shutil
import unittest
from decimal import Decimal
from io import BytesIO
from tempfile import TemporaryDirectory
from time import time
from typing import Optional

from pdfconduit import Pdfconduit, Info
from pdfconduit.utils.typing import Iterable

test_data_dir = os.path.join(os.path.dirname(__file__), "data")
# pdf_name = 'plan_l.pdf'
# pdf_name = 'plan_p.pdf'
# pdf_name = 'article.pdf'
pdf_name = "document.pdf"
# pdf_name = 'con docs2.pdf'
img_name = "floor plan.png"
pdf_path = os.path.join(test_data_dir, pdf_name)
img_path = os.path.join(test_data_dir, img_name)


def test_data_path(filename):
    return os.path.join(test_data_dir, filename)


def files_are_equal(file1_path, file2_path):
    """
    Compare two files to check if they are the same.

    Parameters:
    file1_path (str): The path to the first file.
    file2_path (str): The path to the second file.

    Returns:
    bool: True if the files are the same, False otherwise.
    """
    with open(file1_path, "rb") as file1, open(file2_path, "rb") as file2:
        file1_content = file1.read()
        file2_content = file2.read()

    return file1_content == file2_content


def function_name_to_file_name(extension=".pdf"):
    return inspect.stack()[1][3] + extension


def copy_pdf_to_output_directory(pdf, save_name):
    shutil.copy(pdf, get_output_filepath(save_name))


def get_output_filepath(filename):
    return os.path.join(os.path.join(os.path.dirname(__file__), "output"), filename)


def expected_equals_output(test_function, output_filepath):
    return files_are_equal(get_output_filepath(test_function), output_filepath)


def get_clean_pdf_name(path) -> str:
    return os.path.basename(str(path)).replace(".", ",")


# Example usage
# Save outputs to tests/output directory
# copy_pdf_to_output_directory({pdfoutput}, function_name_to_file_name())

# Compare saved outputs to generate pdfs
# expected_equals_output(function_name_to_file_name(), {pdfouput})


class PdfconduitTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Encryption passwords
        cls.owner_pw = "foo"
        cls.user_pw = "baz"
        cls.pdf_path = pdf_path
        cls.timings = {}

    def setUp(self):
        self.temp = TemporaryDirectory()
        self.conduit = Pdfconduit(self.pdf_path).set_output_directory(self.temp.name)
        self.timer = Timer()

    def tearDown(self):
        self.temp.cleanup()
        self.timings[self.id()] = self.timer.end

    @classmethod
    def tearDownClass(cls):
        json_path = test_data_path("timings.json")

        if os.path.isfile(json_path):
            with open(json_path, "r") as json_file:
                data = json.load(json_file)

            data.update(cls.timings)
        else:
            data = cls.timings

        with open(json_path, "w") as json_file:
            json.dump(data, json_file, indent=2, sort_keys=True)

    def assertPdfExists(self, pdf):
        self.assertTrue(os.path.exists(pdf))
        self.assertTrue(os.path.isfile(pdf))

    def assertPdfDoesntExists(self, pdf):
        self.assertFalse(os.path.exists(pdf))
        self.assertFalse(os.path.isfile(pdf))

    def assertPdfPagesEqual(self, original: str, modified: str):
        info_og = Info(original)
        info_modified = Info(modified)
        self.assertEqual(info_og.pages, info_modified.pages)
        self.assertTrue(abs(info_og.size[0] / info_modified.size[0]) <= 1)
        self.assertTrue(abs(info_og.size[1] / info_modified.size[1]) <= 1)

    def assertCorrectNumPages(
        self, main_pdf: str, pdfs_to_merge: Iterable[str], expected_pages: int
    ) -> None:
        # Assert sum of pages in original pdf files equals sum of pages in merged pdf
        self.assertEqual(
            sum([Info(pdf).pages for pdf in pdfs_to_merge]) + Info(main_pdf).pages,
            expected_pages,
        )

    def assertPdfRotation(self, rotated, rotation):
        info = rotated.info if isinstance(rotated, Pdfconduit) else rotated
        self.assertEqual(info.rotate, rotation)

    def assertCorrectPagesSliced(self, fp, lp, sliced):
        info = sliced.info if isinstance(sliced, Pdfconduit) else sliced
        self.assertEqual(info.pages, len(range(fp, lp + 1)))

    def assertPdfScaled(self, scale, scaled, original=None):
        original = self.pdf_path if original is None else original
        info = scaled.info if isinstance(scaled, Pdfconduit) else scaled
        self.assertEqual(info.size, tuple([i * scale for i in Info(original).size]))
        # self.assertEqual(info.pages, Info(original).pages)

    def assertFileSizeDecreased(self, original: str, modified: str):
        self.assertLess(os.path.getsize(modified), os.path.getsize(original))

    def _get_pdf_byte_stream(self, path: Optional[str] = None) -> BytesIO:
        with open(path if path else self.pdf_path, "rb") as fh:
            return BytesIO(fh.read())


class Timer:
    def __init__(self, decimal_places=2):
        self._decimal_places = decimal_places

        self.start = time()

    @property
    def end(self):
        # Calculate run time
        return self.human_time(time() - self.start)

    @staticmethod
    def rounder(exact, decimals=2):
        """Round a float to a certain number of decimal places."""
        return float(round(Decimal(exact), decimals))

    def human_time(self, runtime):
        """Display runtime in a human friendly format."""
        return self.rounder(runtime * 1000, self._decimal_places)


__all__ = [
    "pdf_path",
    "img_path",
    "test_data_dir",
    "test_data_path",
    "function_name_to_file_name",
    "copy_pdf_to_output_directory",
    "expected_equals_output",
    "get_output_filepath",
    "PdfconduitTestCase",
    "get_clean_pdf_name",
]



from typing import Optional, Any, Tuple, Dict, Union, List, Iterable

try:
    from typing import Self, Annotated
except ImportError:
    from typing_extensions import Self, Annotated


__all__ = [Optional, Any, Tuple, Dict, Self, Annotated, Union, List, Iterable]



# Write two (2) PDFs to a destination file
from pathlib import Path
from typing import Union

from pypdf import PdfReader as PypdfReader, PdfWriter as PypdfWriter
from pypdf._utils import StrByteType


def write_pdf(pdf_obj: Union[StrByteType, Path], destination: str):
    """
    Write PDF object to file
    :param pdf_obj: PDF object to be written to file
    :param destination: Destination path
    """
    reader = PypdfReader(pdf_obj)  # Create new PDF object
    writer = PypdfWriter()

    page_count = reader.get_num_pages()

    # add the "watermark" (which is the new pdf) on the existing page
    for page_number in range(page_count):
        page = reader.get_page(page_number)
        writer.add_page(page)

    # finally, write "output" to a real file
    with open(destination, "wb") as outputStream:
        writer.write(outputStream)



import os
import unittest
from tempfile import TemporaryDirectory

from pdfconduit.convert import IMG2PDF
from tests import *


class TestImg2Pdf(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.img_path = img_path
        cls.pdf = None
        cls.tempdir = TemporaryDirectory()

    @classmethod
    def tearDownClass(cls):
        if hasattr(cls, "tempdir"):
            cls.tempdir.cleanup()

    def tearDown(self):
        if os.path.exists(self.pdf):
            os.remove(self.pdf)

    def test_convert(self):
        """Convert an image file into PDF."""
        ip = IMG2PDF(tempdir=self.tempdir)
        self.pdf = ip.convert(self.img_path)

        # Assert pdf file exists
        self.assertTrue(os.path.exists(self.pdf))

    def test_convert_packet(self):
        """Convert an image file into PDF."""
        self.pdf = IMG2PDF(
            [self.img_path, self.img_path, self.img_path],
            destination=test_data_dir,
            tempdir=self.tempdir,
        ).save(clean_temp=False)

        # Assert pdf file exists
        self.assertTrue(os.path.exists(self.pdf))


if __name__ == "__main__":
    unittest.main()



import os
from typing import Tuple, List

from parameterized import parameterized

from pdfconduit import Info, Pdfconduit
from pdfconduit.settings import Compression
from tests import PdfconduitTestCase, get_clean_pdf_name
from tests import test_data_path


def optimization_params() -> List[str]:
    return list(
        map(
            lambda filename: test_data_path(filename),
            [
                "article.pdf",
                "charts.pdf",
                "document.pdf",
                "workbook.pdf",
            ],
        )
    )


def optimizations_name_func(testcase_func, param_num, param):
    return "{}.{}".format(testcase_func.__name__, get_clean_pdf_name(param.args[0]))


def compress_params() -> List[Tuple[str, Compression]]:
    return [
        (filepath, level)
        for filepath in optimization_params()[0:3]
        for level in Compression.all()
    ]


def compress_name_func(testcase_func, param_num, param):
    return "{}.{}.level_{}".format(
        testcase_func.__name__,
        get_clean_pdf_name(param.args[0]),
        param.args[1].value,
    )


class TestOptimizations(PdfconduitTestCase):
    @parameterized.expand(optimization_params, name_func=optimizations_name_func)
    def test_minify(self, pdf_path: str):
        self.conduit = Pdfconduit(pdf_path).set_output_directory(self.temp.name)
        self.conduit.minify().write()

        self.assertPdfExists(self.conduit.output)
        self.assertFileSizeDecreased(pdf_path, self.conduit.output)
        self.assertPdfPagesEqual(pdf_path, self.conduit.output)

    @parameterized.expand(optimization_params, name_func=optimizations_name_func)
    def test_remove_duplication(self, pdf_path: str):
        self.conduit = Pdfconduit(pdf_path).set_output_directory(self.temp.name)
        self.conduit.remove_duplication().write()

        self.assertPdfExists(self.conduit.output)
        self.assertFileSizeDecreased(pdf_path, self.conduit.output)
        self.assertPdfPagesEqual(pdf_path, self.conduit.output)

    @parameterized.expand(optimization_params, name_func=optimizations_name_func)
    def test_remove_images(self, pdf_path: str):
        self.conduit = Pdfconduit(pdf_path).set_output_directory(self.temp.name)
        self.conduit.remove_images().write()

        self.assertPdfExists(self.conduit.output)
        self.assertEqual(0, self.conduit.info.images_count)

    @parameterized.expand(optimization_params, name_func=optimizations_name_func)
    def test_reduce_image_quality(self, pdf_path: str):
        if pdf_path.endswith("workbook.pdf"):
            self.skipTest("workbook.pdf cannot be reduced quality")
        self.conduit = Pdfconduit(pdf_path).set_output_directory(self.temp.name)
        self.conduit.reduce_image_quality(50).write()

        self.assertPdfExists(self.conduit.output)
        self.assertEqual(Info(pdf_path).images_count, self.conduit.info.images_count)
        self.assertFileSizeDecreased(pdf_path, self.conduit.output)

    @parameterized.expand(compress_params, name_func=compress_name_func)
    def test_compress(self, pdf_path: str, compression: Compression):
        self.conduit = Pdfconduit(pdf_path).set_output_directory(self.temp.name)
        self.conduit.compress(compression).write()

        self.assertPdfExists(self.conduit.output)
        self.assertEqual(Info(pdf_path).images_count, self.conduit.info.images_count)
        # todo: fix compression increasing size, maybe newer pdfs?
        # self.assertFileSizeDecreased(pdf_path, self.conduit.output)



from typing import List

from parameterized import parameterized

from pdfconduit import Pdfconduit
from tests import PdfconduitTestCase, get_clean_pdf_name
from tests import test_data_path


def flatten_params() -> List[str]:
    return list(
        map(
            lambda filename: test_data_path(filename),
            [
                "article.pdf",
                "charts.pdf",
                "document.pdf",
            ],
        )
    )


def flatten_name_func(testcase_func, param_num, param):
    return "{}.{}".format(testcase_func.__name__, get_clean_pdf_name(param.args[0]))


class TestFlatten(PdfconduitTestCase):
    @parameterized.expand(flatten_params, name_func=flatten_name_func)
    def test_flatten(self, pdf_path: str):
        self.conduit = Pdfconduit(pdf_path).set_output_directory(self.temp.name)
        self.conduit.flatten().write()

        self.assertPdfExists(self.conduit.output)
        self.assertPdfPagesEqual(pdf_path, self.conduit.output)
        # todo: improve assertions

    @parameterized.expand(flatten_params, name_func=flatten_name_func)
    def test_flatten_from_stram(self, pdf_path: str):
        self.conduit = Pdfconduit(
            self._get_pdf_byte_stream(pdf_path)
        ).set_output_directory(self.temp.name)
        self.conduit.flatten().write()

        self.assertPdfExists(self.conduit.output)
        self.assertPdfPagesEqual(pdf_path, self.conduit.output)
        # todo: improve assertions



# Set directory paths and file names
import os
from pathlib import Path
from typing import Optional


def add_suffix(
    file_path: str, suffix: str = "modified", sep: str = "_", ext: Optional[str] = None
):
    """Adds suffix to a file name seperated by an underscore and returns file path."""
    p = Path(file_path)
    _ext = p.suffix if ext is None else str("." + ext.strip("."))
    out = p.stem + sep + suffix + _ext  # p.suffix is file extension
    return os.path.join(os.path.dirname(file_path), out)



# Retrieve information about a PDF document
from typing import Optional, List, Dict, Tuple, Union

from pypdf import PdfReader, DocumentInformation, PageObject, PdfWriter

from pdfconduit.utils._permissions import Permissions
from pdfconduit.utils.read import pypdf_reader


class Info:
    def __init__(
        self, pdf: Union[str, PdfReader, PdfWriter], password: Optional[str] = None
    ):
        self.pdf = (
            pdf
            if isinstance(pdf, (PdfReader, PdfWriter))
            else pypdf_reader(pdf, password)
        )

    @staticmethod
    def _resolved_objects(pdf: PdfReader, xobject: str) -> int:
        """Retrieve rotation info."""
        return [pdf.get_page(i).get(xobject) for i in range(pdf.get_num_pages())][0]

    @property
    def encrypted(self) -> bool:
        """Check weather a PDF is encrypted"""
        return self.pdf.is_encrypted

    @property
    def decrypted(self) -> bool:
        """Check weather a PDF is encrypted"""
        return not self.encrypted

    @property
    def pages(self) -> int:
        """Retrieve PDF number of pages"""
        return self.pdf.get_num_pages()

    @property
    def metadata(self) -> Optional[DocumentInformation]:
        """Retrieve PDF metadata"""
        return self.pdf.metadata

    def resources(self) -> List[PageObject]:
        """Retrieve contents of each page of PDF"""
        # todo: refactor to generator?
        return [self.pdf.get_page(i) for i in range(self.pdf.get_num_pages())]

    @property
    def security(self) -> Dict[str, int]:
        """Print security object information for a pdf document"""
        return {
            k: v for i in self.pdf.resolved_objects.items() for k, v in i[1].items()
        }

    @property
    def dimensions(self) -> Dict[str, float]:
        """Get width and height of a PDF"""
        # todo: add page parameter?
        # todo: add height & width methods?
        # todo: add typed dict
        size = self.pdf.get_page(0).mediabox
        return {"w": float(size[2]), "h": float(size[3])}

    @property
    def size(self) -> Tuple[float, float]:
        """Get width and height of a PDF"""
        size = self.pdf.get_page(0).mediabox

        return float(size[2]), float(size[3])

    @property
    def rotate(self) -> int:
        """Retrieve rotation info."""
        # todo: add page param
        # todo: refactor to `rotation()`
        # todo: add is_rotated
        return self._resolved_objects(self.pdf, "/Rotate")

    @property
    def permissions(self) -> Permissions:
        """Retrieve user access permissions."""
        return Permissions(self.pdf)

    @property
    def images_count(self) -> int:
        return sum([len(page.images) for page in self.pdf.pages])



from datetime import datetime
from typing import Optional, Union, Tuple

from PillowImage import PillowImage

from pdfconduit.watermark.modify.canvas.objects import (
    CanvasObjects,
    CanvasStr,
    CanvasImg,
)


class CanvasConstructor:
    def __init__(
        self,
        text1: str = None,
        text2: str = None,
        copyright_: Optional[bool] = None,
        image: Optional[str] = None,
        rotate: int = 0,
        opacity: Union[int, float] = 8,
        tempdir: Optional[str] = None,
    ):
        self.text1 = text1
        self.text2 = text2
        self.copyright = copyright_
        self.image = image
        self.rotate = rotate
        self.opacity = opacity
        self.tempdir = tempdir

        # Initialize CanvasObjects collector class and add objects
        self.obj = CanvasObjects()

    @property
    def objects(self) -> Tuple[CanvasObjects, int]:
        return self.obj, self.rotate

    def canvas(self) -> Tuple[CanvasObjects, int]:
        if self.image is not None:
            self.obj.add(CanvasImg(self.image, opacity=self.opacity, x=0, y=253))

            if self.text1 and self.text2 and self.copyright:
                self.obj.add(
                    CanvasStr(
                        "© copyright " + str(datetime.now().year),
                        size=16,
                        y=-170,
                        opacity=self.opacity,
                    )
                )
                self.obj.add(
                    CanvasStr(self.text1, opacity=self.opacity, size=40, y=-360)
                )
                self.obj.add(
                    CanvasStr(self.text2, opacity=self.opacity, size=40, y=-560)
                )
            elif self.text1 and self.text2 and not self.copyright:
                self.obj.add(CanvasStr(self.text1, opacity=self.opacity, size=40, y=90))
                self.obj.add(CanvasStr(self.text2, opacity=self.opacity, size=40, y=0))
            elif self.text1:
                self.obj.add(
                    CanvasStr(self.text1, opacity=self.opacity, size=40, y=-125)
                )
            elif self.copyright:
                self.obj.add(
                    CanvasStr(self.text1, opacity=self.opacity, size=40, y=-125)
                )

        else:
            if self.text1 and self.text2 and self.copyright:
                self.obj.add(
                    CanvasStr(self.text1, opacity=self.opacity, size=80, y=110)
                )
                self.obj.add(CanvasStr(self.text2, opacity=self.opacity, size=80, y=20))
                self.obj.add(
                    CanvasStr(
                        "© copyright " + str(datetime.now().year),
                        size=16,
                        y=-30,
                        opacity=self.opacity,
                    )
                )
            elif self.text1 and self.text2 and not self.copyright:
                self.obj.add(CanvasStr(self.text1, opacity=self.opacity, size=80, y=90))
                self.obj.add(CanvasStr(self.text2, opacity=self.opacity, size=80, y=0))
            elif self.text1:
                self.obj.add(
                    CanvasStr(self.text1, opacity=self.opacity, size=40, y=-125)
                )
            elif self.copyright:
                self.obj.add(
                    CanvasStr(
                        "© copyright " + str(datetime.now().year),
                        opacity=self.opacity,
                        size=16,
                        y=50,
                    )
                )
        return self.objects

    def img(self) -> Tuple[CanvasObjects, int]:
        with PillowImage() as img:
            if self.image is not None:
                img.draw_img(self.image, x=0, y=50, opacity=self.opacity)

                if self.text1 and self.text2 and self.copyright:
                    img.draw_text(self.text1, font_size=40, y=416, opacity=self.opacity)
                    img.draw_text(self.text2, font_size=40, y=466, opacity=self.opacity)
                    img.draw_text(
                        "© copyright " + str(datetime.now().year),
                        font_size=16,
                        y="center",
                    )
                elif self.text1 and self.text2 and not self.copyright:
                    img.draw_text(self.text1, font_size=40, y=416, opacity=self.opacity)
                    img.draw_text(self.text2, font_size=40, y=466, opacity=self.opacity)
                elif self.text1:
                    self.obj.add(
                        CanvasStr(self.text1, opacity=self.opacity, size=40, y=-125)
                    )
                elif self.copyright:
                    img.draw_text(
                        "© copyright " + str(datetime.now().year),
                        font_size=16,
                        y="center",
                    )
            else:
                if self.text1 and self.text2 and self.copyright:
                    img.draw_text(
                        "© copyright " + str(datetime.now().year), font_size=16, y=350
                    )
                    img.draw_text(self.text1, font_size=80, y=106, opacity=self.opacity)
                    img.draw_text(self.text2, font_size=80, y=206, opacity=self.opacity)
                elif self.text1 and self.text2 and not self.copyright:
                    img.draw_text(self.text1, font_size=80, y=106, opacity=self.opacity)
                    img.draw_text(self.text2, font_size=80, y=206, opacity=self.opacity)
                elif self.text1:
                    img.draw_text(self.text1, font_size=80, y=106, opacity=self.opacity)
                elif self.copyright:
                    img.draw_text(
                        "© copyright " + str(datetime.now().year), font_size=16, y=350
                    )

            img.rotate(self.rotate)
            self.rotate = 0
            i = img.save(destination=self.tempdir)
            self.obj.add(CanvasImg(i, opacity=1, centered=True))
            return self.objects



import os
from typing import List, Tuple

from parameterized import parameterized

from pdfconduit import Info
from pdfconduit.settings import Algorithms, Encryption
from tests import PdfconduitTestCase


def encryption_name_func(testcase_func, param_num, param):
    name = ""
    if not param.args[0].allow_printing and not param.args[0].allow_commenting:
        name += "_none"
    if param.args[0].allow_printing:
        name += "_printing"
    if param.args[0].allow_commenting:
        name += "_commenting"
    return "{}.{}.{}".format(testcase_func.__name__, param.args[0].algo.value, name[1:])


def encryption_params() -> List[Tuple[Encryption, int]]:
    algos = [
        # (name, algo, expected_security_handler)
        (Algorithms.RC4_40, 2),
        (Algorithms.RC4_128, 3),
        (Algorithms.AES_128, 4),
        (Algorithms.AES_256, 6),
        (Algorithms.AES_256_r5, 5),
    ]

    permissions = [
        {"allow_printing": True, "allow_commenting": False},
        {"allow_printing": True, "allow_commenting": True},
        {"allow_printing": False, "allow_commenting": True},
        {"allow_printing": False, "allow_commenting": False},
    ]

    params = []
    for algo in algos:
        algorithm, expected_security_handler = algo
        for permission in permissions:
            params.append(
                (
                    Encryption(
                        user_pw="baz",
                        owner_pw="foo",
                        allow_printing=permission["allow_printing"],
                        allow_commenting=permission["allow_commenting"],
                        algo=algorithm,
                    ),
                    expected_security_handler,
                )
            )
    return params


class EncryptionTestCase(PdfconduitTestCase):
    def _getPdfSecurity(self, encrypted):
        return Info(encrypted.output, self.user_pw).security

    def assertEncrypted(self, pdf):
        self.assertTrue(Info(pdf.output, self.user_pw).encrypted)

    def assert40BitEncryption(self, security):
        if "/Length" in security:
            self.assertEqual(security["/Length"], 40)

        # Assert standard security handler revision is 2
        self.assertEqual(security["/R"], 2)

    def assert128BitEncryption(self, security, security_handler_revision: int = 4):
        self.assertTrue("/Length" in security)
        self.assertIsInstance(security["/Length"], int)
        self.assertEqual(security["/Length"], 128)

        self.assertEqual(security["/Filter"], "/Standard")

        # Assert standard security handler revision is 3
        self.assertEqual(security["/R"], security_handler_revision)

        if security_handler_revision == 4:
            self.assertTrue("/CF" in security)
            self.assertEqual(security["/CF"]["/StdCF"]["/CFM"], "/AESV2")
        else:
            self.assertFalse("/CF" in security)

    def assert256BitEncryption(self, security, security_handler_revision: int = 5):
        self.assertTrue("/Length" in security)
        self.assertIsInstance(security["/Length"], int)
        self.assertEqual(security["/Length"], 256)

        self.assertEqual(security["/Filter"], "/Standard")

        # Assert standard security handler revision is 3
        self.assertEqual(security["/R"], security_handler_revision)
        self.assertEqual(security["/V"], 5)

        self.assertTrue("/CF" in security)
        self.assertEqual(security["/CF"]["/StdCF"]["/CFM"], "/AESV3")

    def assertSecurityValue(self, security, expected):
        self.assertTrue("/P" in security)
        self.assertEqual(security["/P"], expected)

    def assertPermissions(
        self,
        pdf,
        can_print=False,
        can_modify=False,
        can_copy=False,
        can_annotate=False,
        can_fill_forms=False,
        can_change_accessability=False,
        can_assemble=False,
        can_print_high_quality=False,
    ):
        permissions = Info(pdf.output, self.user_pw).permissions
        self.assertEqual(permissions.can_print(), can_print)
        self.assertEqual(permissions.can_modify(), can_modify)
        self.assertEqual(permissions.can_copy(), can_copy)
        self.assertEqual(permissions.can_annotate(), can_annotate)
        self.assertEqual(permissions.can_fill_forms(), can_fill_forms)
        self.assertEqual(
            permissions.can_change_accessability(), can_change_accessability
        )
        self.assertEqual(permissions.can_assemble(), can_assemble)
        self.assertEqual(permissions.can_print_high_quality(), can_print_high_quality)


class TestEncryption(EncryptionTestCase):
    @parameterized.expand(encryption_params, name_func=encryption_name_func)
    def test_can_encrypt_pdf(
        self, encryption: Encryption, expected_security_handler: int
    ):
        self.conduit.encrypt(encryption).write()

        security = self._getPdfSecurity(self.conduit)

        self.assertPdfExists(self.conduit.output)
        self.assertEncrypted(self.conduit)
        if encryption.algo.bit_length == 40:
            self.assert40BitEncryption(security)
        if encryption.algo.bit_length == 128:
            self.assert128BitEncryption(security, expected_security_handler)
        if encryption.algo.bit_length == 256:
            self.assert256BitEncryption(security, expected_security_handler)

        # self.assertSecurityValue(security, 4)
        self.assertPermissions(
            self.conduit,
            can_print=encryption.allow_printing,
            can_modify=encryption.allow_commenting,
        )

    def test_password_byte_string(self):
        self.conduit.encrypt(
            Encryption(user_pw="baz", owner_pw="foo", algo=Algorithms.RC4_128)
        ).write()

        security = self._getPdfSecurity(self.conduit)

        self.assertPdfExists(self.conduit.output)
        self.assertEncrypted(self.conduit)
        self.assert128BitEncryption(security, 3)
        self.assertSecurityValue(security, 4)

        self.assertPermissions(self.conduit, can_print=True)

        # Assert user & owner password byte string is correct
        self.assertEqual(
            security["/O"],
            "ÐHÑRžS]˘—Í8V6{˘KJ’ß\x01æ7ÑnÊ\x06[¯Nd’\x0b",
        )



# Convert each page of PDF to images
import os
from io import BytesIO
from tempfile import NamedTemporaryFile
from typing import Optional, List

import fitz
from PIL import Image

from pdfconduit.utils.path import add_suffix


class PDF2IMG:
    def __init__(
        self,
        file_name: str,
        output: Optional[str] = None,
        tempdir: Optional[str] = None,
        ext: str = ".png",
        alpha: bool = False,
    ):
        """Convert each page of a PDF file into a PNG image"""
        self.file_name = file_name
        self.output = output
        self.tempdir = tempdir
        self.ext = ext
        self.alpha = alpha

        self.doc = fitz.open(self.file_name)
        self.output_dir = os.path.dirname(file_name) if tempdir is None else tempdir

        # storage for page display lists
        self.dlist_tab = [None] * len(self.doc)
        self._page_data = None

    @property
    def pdf_data(self) -> List[bytes]:
        if not self._page_data:
            self._page_data = self._get_pdf_data()
        return self._page_data

    def _get_pdf_data(self) -> List[bytes]:
        return [self._get_page_data(cur_page) for cur_page in range(len(self.doc))]

    def _get_page_data(self, pno, zoom=0) -> bytes:
        """
        Return a PNG image for a document page number. If zoom is other than 0, one of
        the 4 page quadrants are zoomed-in instead and the corresponding clip returned.
        """
        dlist = self.dlist_tab[pno]  # get display list
        if not dlist:  # create if not yet there
            self.dlist_tab[pno] = self.doc[pno].get_displaylist()
            dlist = self.dlist_tab[pno]
        r = dlist.rect  # page rectangle
        mp = r.tl + (r.br - r.tl) * 0.5  # rect middle point
        mt = r.tl + (r.tr - r.tl) * 0.5  # middle of top edge
        ml = r.tl + (r.bl - r.tl) * 0.5  # middle of left edge
        mr = r.tr + (r.br - r.tr) * 0.5  # middle of right egde
        mb = r.bl + (r.br - r.bl) * 0.5  # middle of bottom edge
        mat = fitz.Matrix(2, 2)  # zoom matrix
        if zoom == 1:  # top-left quadrant
            clip = fitz.Rect(r.tl, mp)
        elif zoom == 4:  # bot-right quadrant
            clip = fitz.Rect(mp, r.br)
        elif zoom == 2:  # top-right
            clip = fitz.Rect(mt, mr)
        elif zoom == 3:  # bot-left
            clip = fitz.Rect(ml, mb)
        if zoom == 0:  # total page
            pix = dlist.get_pixmap(alpha=self.alpha)
        else:
            pix = dlist.get_pixmap(alpha=self.alpha, matrix=mat, clip=clip)
        return pix.tobytes()  # return the PNG image

    def _get_output(self, index: int) -> str:
        if self.output:
            return self.output
        elif not self.tempdir:
            output_file = add_suffix(self.file_name, str(index + 1), ext=self.ext)
            return os.path.join(self.output_dir, output_file)
        else:
            with NamedTemporaryFile(
                suffix=self.ext, dir=self.tempdir, delete=True
            ) as temp:
                return temp.name

    def save(self) -> List[str]:
        saved = []
        for i, img in enumerate(self.pdf_data):
            output = self._get_output(i)
            saved.append(output)
            with Image.open(BytesIO(img)) as image:
                image.save(output)
        self.doc.close()
        return saved



import os
from typing import Tuple, List, Optional

from parameterized import parameterized

from pdfconduit import Info
from pdfconduit.utils import add_suffix
from tests import PdfconduitTestCase, get_clean_pdf_name
from tests import test_data_path


def unencrypted_pdf_params() -> List[str]:
    return list(
        map(
            lambda filename: test_data_path(filename),
            ["article.pdf", "manual.pdf", "workbook.pdf"],
        )
    )


def encrypted_pdf_params() -> List[str]:
    return list(
        map(
            lambda filepath: add_suffix(filepath, "encrypted"), unencrypted_pdf_params()
        )
    )


def pages_params() -> List[Tuple[str, int]]:
    return [
        (test_data_path("article.pdf"), 1),
        (test_data_path("manual.pdf"), 82),
        (test_data_path("workbook.pdf"), 128),
    ]


def dimensions_params() -> List[Tuple[str, float, float]]:
    default_w = 595.276
    default_h = 841.89
    return [
        (test_data_path("article.pdf"), default_w, default_h),
        (test_data_path("manual.pdf"), default_w, default_h),
        (test_data_path("workbook.pdf"), default_w, default_h),
        (test_data_path("con docs.pdf"), 2592.0, 1728.0),
    ]


def info_name_func(testcase_func, param_num, param):
    return "{}.{}".format(
        testcase_func.__name__, get_clean_pdf_name(param.args[0]).replace(" ", "_")
    )


class TestInfo(PdfconduitTestCase):
    @parameterized.expand(encrypted_pdf_params, name_func=info_name_func)
    def test_is_encrypted(self, filepath: str):
        info = self._get_info(filepath)
        self.assertTrue(info.encrypted)

    @parameterized.expand(unencrypted_pdf_params, name_func=info_name_func)
    def test_is_decrypted(self, filepath: str):
        info = self._get_info(filepath)
        self.assertTrue(info.decrypted)

    @parameterized.expand(unencrypted_pdf_params, name_func=info_name_func)
    def test_is_not_encrypted(self, filepath: str):
        info = self._get_info(filepath)

        self.assertFalse(info.encrypted)

    @parameterized.expand(encrypted_pdf_params, name_func=info_name_func)
    def test_is_not_decrypted(self, filepath: str):
        info = self._get_info(filepath)
        self.assertFalse(info.decrypted)

    @parameterized.expand(pages_params, name_func=info_name_func)
    def test_pages(self, filepath: str, expected_pages: int):
        info = self._get_info(filepath)

        self.assertIsInstance(info.pages, int)
        self.assertEqual(info.pages, expected_pages)

    def test_metadata(self):
        info = self._get_info(test_data_path("article.pdf"))

        self.assertIsInstance(info.metadata, dict)
        self.assertEqual(
            info.metadata["/Creator"], "This PDF is created by PDF4U Pro 2.0"
        )
        self.assertEqual(
            info.metadata["/CreationDate"],
            "D:20040120105826",
        )
        self.assertEqual(
            info.metadata["/Producer"],
            "PDF4U Adobe PDF Creator 2.0",
        )

    def test_resources(self):
        info = self._get_info(test_data_path("article.pdf"))

        self.assertEqual(info.pages, len(info.resources()))

        resources = info.resources()[0]
        self.assertIsInstance(resources, dict)
        self.assertEqual(resources["/Type"], "/Page")

        mediabox = list(resources["/MediaBox"])
        self.assertEqual(mediabox[0], 0)
        self.assertEqual(mediabox[1], 0)
        self.assertEqual(float(mediabox[2]), 595.276)
        self.assertEqual(float(mediabox[3]), 841.89)

    @parameterized.expand(encrypted_pdf_params, name_func=info_name_func)
    def test_security_encrypted_pdf(self, filepath: str):
        info = self._get_info(filepath)

        self.assertIsInstance(info.security, dict)
        self.assertTrue("/V" in info.security)
        self.assertTrue("/R" in info.security)
        self.assertTrue("/Length" in info.security)
        self.assertTrue("/P" in info.security)
        self.assertTrue("/Filter" in info.security)
        self.assertTrue("/O" in info.security)
        self.assertTrue("/U" in info.security)

    @parameterized.expand(encrypted_pdf_params, name_func=info_name_func)
    def test_security_decrypted_pdf(self, filepath: str):
        info = self._get_info(filepath, "baz")

        self.assertIsInstance(info.security, dict)
        self.assertTrue("/V" in info.security)
        self.assertTrue("/R" in info.security)
        self.assertTrue("/Length" in info.security)
        self.assertTrue("/P" in info.security)
        self.assertTrue("/Filter" in info.security)
        self.assertTrue("/O" in info.security)
        self.assertTrue("/U" in info.security)

    @parameterized.expand(unencrypted_pdf_params, name_func=info_name_func)
    def test_security_passwordless_pdf(self, filepath: str):
        info = self._get_info(filepath)

        self.assertIsInstance(info.security, dict)
        self.assertEqual(info.security, {})

    @parameterized.expand(dimensions_params, name_func=info_name_func)
    def test_dimensions(
        self, filepath: str, expected_width: float, expected_height: float
    ):
        info = self._get_info(filepath)

        self.assertIsInstance(info.dimensions, dict)
        self.assertTrue("w" in info.dimensions)
        self.assertTrue("h" in info.dimensions)
        self.assertIsInstance(info.dimensions["w"], float)
        self.assertIsInstance(info.dimensions["h"], float)
        self.assertEqual(info.dimensions["w"], expected_width)
        self.assertEqual(info.dimensions["h"], expected_height)

    @parameterized.expand(dimensions_params, name_func=info_name_func)
    def test_size(self, filepath: str, expected_width: float, expected_height: float):
        info = self._get_info(filepath)

        self.assertIsInstance(info.size, tuple)
        self.assertEqual(len(info.size), 2)
        self.assertIsInstance(info.size[0], float)
        self.assertIsInstance(info.size[1], float)
        self.assertEqual(info.size[0], expected_width)
        self.assertEqual(info.size[1], expected_height)

    @parameterized.expand(dimensions_params, name_func=info_name_func)
    def test_size_and_dimensions_are_equal(
        self, filepath: str, expected_width: float, expected_height: float
    ):
        info = self._get_info(filepath)
        self.assertEqual(info.size[0], info.dimensions["w"])
        self.assertEqual(info.size[1], info.dimensions["h"])

    @parameterized.expand(unencrypted_pdf_params, name_func=info_name_func)
    def test_rotate_no_rotation(self, filepath: str):
        info = self._get_info(filepath)
        self.assertEqual(info.rotate, None)

    def test_rotate_rotated(self):
        info = self._get_info(test_data_path("rotated.pdf"))

        self.assertIsInstance(info.rotate, int)
        self.assertEqual(info.rotate, 90)

    @staticmethod
    def _get_info(filepath: str, password: Optional[str] = None):
        return Info(filepath, password=password)


def get_expected_output(filepath: str, suffix: str = "modified", sep: str = "_") -> str:
    return os.path.join(
        os.path.dirname(filepath),
        os.path.basename(filepath).replace(".pdf", "") + sep + suffix + ".pdf",
    )


def path_params() -> List[Tuple[str, str, str]]:
    return [
        (get_expected_output(path, suffix), path, suffix)
        for path in unencrypted_pdf_params()
        for suffix in ["modified", "new", "old", "backup", "changed"]
    ]


def path_params_with_sep() -> List[Tuple[str, str, str, str]]:
    return [
        (get_expected_output(path, suffix, sep), path, suffix, sep)
        for path in unencrypted_pdf_params()
        for suffix in ["modified", "new", "old", "backup", "changed"]
        for sep in ["_", "-"]
    ]


def path_name_func(testcase_func, param_num, param):
    return "{}-{}".format(
        testcase_func.__name__, os.path.basename(str(param.args[0])).replace(" ", "_")
    )


class TestPath(PdfconduitTestCase):
    @parameterized.expand(path_params, name_func=info_name_func)
    def test_add_suffix(self, expected: str, filepath: str, suffix: str):
        with_suffix = add_suffix(filepath, suffix)

        self.assertIsInstance(with_suffix, str)
        self.assertEqual(with_suffix, expected)

    @parameterized.expand(path_params_with_sep, name_func=info_name_func)
    def test_add_suffix_suffix_sep(
        self, expected: str, filepath: str, suffix: str, sep: str
    ):
        with_suffix = add_suffix(filepath, suffix, sep)

        self.assertIsInstance(with_suffix, str)
        self.assertEqual(with_suffix, expected)

    @parameterized.expand(path_params_with_sep, name_func=info_name_func)
    def test_add_suffix_suffix_sep_ext(
        self, expected: str, filepath: str, suffix: str, sep: str
    ):
        with_suffix = add_suffix(filepath, suffix, sep, "zip")

        self.assertIsInstance(with_suffix, str)
        self.assertEqual(with_suffix, expected.replace(".pdf", ".zip"))



import os
import warnings
from io import BufferedReader

from pypdf import PdfReader, PdfWriter

from pdfconduit import Pdfconduit
from pdfconduit.utils import add_suffix
from tests import *
from tests import PdfconduitTestCase


class TestUsage(PdfconduitTestCase):
    def test_can_use_context_manager(self):
        with Pdfconduit(self.pdf_path) as conduit:
            conduit.set_output_directory(self.temp.name)

            self.assertIsInstance(conduit._pdf_file, BufferedReader)
            self.assertIsInstance(conduit._reader, PdfReader)
            self.assertIsInstance(conduit._writer, PdfWriter)
            self.assertIsNone(conduit.output)

        self.assertIsInstance(conduit.output, str)
        self.assertTrue(os.path.exists(conduit.output))

    def test_can_read_unencrypted_pdf(self):
        conduit = Pdfconduit(self.pdf_path)

        self.assertIsInstance(conduit._pdf_file, BufferedReader)
        self.assertIsInstance(conduit._reader, PdfReader)

    def test_can_read_encrypted_pdf(self):
        conduit = Pdfconduit(test_data_path("encrypted.pdf"), self.user_pw)

        self.assertIsInstance(conduit._pdf_file, BufferedReader)
        self.assertIsInstance(conduit._reader, PdfReader)

    def test_can_set_default_metadata(self):
        original = self.conduit.info.metadata

        self.assertNotEqual("pdfconduit", original.producer)
        self.assertNotEqual("pdfconduit", original.creator)
        self.assertNotEqual("pdfconduit", original.author)

        self.conduit.write()

        self.assertNotEqual(original, self.conduit.info.metadata)
        self.assertEqual("pdfconduit", self.conduit.info.metadata.producer)
        self.assertEqual("pdfconduit", self.conduit.info.metadata.creator)
        self.assertEqual("pdfconduit", self.conduit.info.metadata.author)

    def test_can_set_custom_metadata(self):
        original = self.conduit.info.metadata
        self.conduit.set_metadata({"/Producer": "Big Dom"})
        self.conduit.write()

        self.assertNotEqual(original, self.conduit.info.metadata)
        self.assertEqual("Big Dom", self.conduit.info.metadata.producer)

    def test_can_set_output(self):
        self.assertEqual(self.conduit._output_dir, self.temp.name)
        self.conduit.write()
        self.assertEqual(os.path.dirname(self.conduit.output), self.temp.name)

    def test_can_set_output_suffix(self):
        self.conduit = Pdfconduit(self.pdf_path)
        self.conduit.set_output_suffix("changed")
        output = add_suffix(self.pdf_path, "changed")
        self.assertEqual(output, self.conduit.output)

    def test_can_read_from_stream(self):
        conduit = Pdfconduit(self._get_pdf_byte_stream())

        self.assertIsNone(conduit._pdf_file)
        self.assertIsInstance(conduit._reader, PdfReader)
        self.assertIsInstance(conduit._writer, PdfWriter)

    def test_can_read_from_stream_and_write_to_file(self):
        output = os.path.join(self.temp.name, "streamed.pdf")
        conduit = Pdfconduit(self._get_pdf_byte_stream()).set_output(output)

        self.assertIsNone(conduit._pdf_file)
        self.assertIsInstance(conduit._reader, PdfReader)
        self.assertIsInstance(conduit._writer, PdfWriter)

        conduit.write()

        self.assertPdfExists(output)

    def test_can_write_stream_to_file_without_output(self):
        conduit = Pdfconduit(self._get_pdf_byte_stream())

        self.assertIsNone(conduit._pdf_file)
        self.assertIsInstance(conduit._reader, PdfReader)
        self.assertIsInstance(conduit._writer, PdfWriter)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            conduit.write()

            self.assertPdfExists(conduit.output)

            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
            assert "Saving PDFs to a temporary directory because" in str(w[-1].message)

    def test_can_set_temp_output(self):
        self.conduit = Pdfconduit(self.pdf_path)
        self.conduit.set_output_temp()
        self.conduit.write()
        self.assertPdfExists(self.conduit.output)

        self.conduit.cleanup()
        self.assertPdfDoesntExists(self.conduit.output)



from tempfile import TemporaryFile, NamedTemporaryFile

from pypdf import PdfWriter

from pypdf import PdfWriter

from pdfconduit.convert import Flatten
from pdfconduit.internals import BaseConduit
from pdfconduit.settings import Compression, ImageQualityRange, Encryption
from pdfconduit.transform import Merge2
from pdfconduit.transform import Rotate, Upscale
from pdfconduit.utils import Info
from pdfconduit.utils.typing import Optional, Tuple, Self, Annotated


class Pdfconduit(BaseConduit):
    def encrypt(self, encrypter: Encryption) -> Self:
        self._set_default_output("encrypted")
        self._writer.encrypt(
            user_password=encrypter.user_pw,
            owner_password=encrypter.owner_pw,
            permissions_flag=encrypter.permissions,
            algorithm=encrypter.algo.value,
        )
        return self

    def merge(self, pdf: str, position: Optional[int] = None) -> Self:
        # todo: allow for iterable of pdfs
        self._set_default_output("merged")
        if position is None:
            self._writer.append(pdf)
        else:
            self._writer.merge(position, pdf)
        return self

    def merge_fast(self, pdfs: list) -> Self:
        self._set_default_output("merged")
        pdf_objects = [self._stream if self._stream is not None else self._path] + pdfs
        self._path = Merge2(pdf_objects, output=self.output).use_pdfrw().merge()
        return self._open_and_read()

    def rotate(self, degrees: int) -> Self:
        self._set_default_output("rotated")
        for page in self._writer.pages:
            page.rotate(degrees)
        return self

    def rotate_exact(self, degrees: int) -> Self:
        if degrees % 90 == 0:
            return self.rotate(degrees)

        self._set_default_output("rotated")
        self._path = (
            Rotate(
                self._stream if self._stream is not None else self._path,
                degrees,
                output=self.output,
            )
            .use_pdfrw()
            .rotate()
        )
        return self._open_and_read()

    def slice(self, start: int, end: int) -> Self:
        self._set_default_output("sliced")
        start = start - 1  # Reindex page selections for simple user input
        writer = PdfWriter()
        for page_num in list(range(self._writer.get_num_pages()))[start:end]:
            writer.add_page(self._writer.get_page(page_num))
        self._writer = writer
        return self

    def scale(
        self, scale: float, margins: Tuple[int, int] = (0, 0), accelerate: bool = False
    ) -> Self:
        self._set_default_output("scaled")

        if accelerate or margins != (0, 0):
            x, y = margins
            self._path = (
                Upscale(
                    self._path,
                    margin_x=x,
                    margin_y=y,
                    scale=scale,
                    tempdir=self._output_dir,
                )
                .use_pdfrw()
                .upscale()
            )
            return self._open_and_read()

        width, height = self.info.size
        width, height = (width * scale, height * scale)
        for page in self._writer.pages:
            page.scale_to(width, height)
        return self

    def flatten(self) -> Self:
        # todo: re-write Flatten & other convert classes
        # todo: fix issue with flattened pdf output path
        if not self._path and self._stream:
            temp = NamedTemporaryFile(suffix=".pdf")
            temp.write(self._stream.getvalue())
            path = temp.name
        else:
            temp = None
            path = self._path

        if not self._closed:
            self.write()

        self._path = Flatten(path, suffix="flattened", tempdir=self._output_dir).save()

        if temp is not None:
            temp.close()

        return self._open_and_read()

    def minify(self) -> Self:
        # remove duplication (images or pages) from a PDF
        self._set_default_output("minified")
        writer = PdfWriter()
        for page in self._writer.pages:
            writer.add_page(page)
        self._writer = writer
        return self

    def remove_duplication(self) -> Self:
        return self.minify()

    def remove_images(self) -> Self:
        self._set_default_output("noimages")
        self._writer.remove_images()
        return self

    def reduce_image_quality(self, quality: Annotated[int, ImageQualityRange]) -> Self:
        self._set_default_output("reduced")
        for page in self._writer.pages:
            for img in page.images:
                img.replace(img.image, quality=quality)
        return self

    def compress(self, compression: Compression = Compression.DEFAULT) -> Self:
        self._set_default_output("compress_{}".format(compression.value))
        for page in self._writer.pages:
            page.compress_content_streams(compression.value)
        return self

    @property
    def info(self) -> Info:
        return Info(self._writer if not self._closed else self.output)

    def watermark(self):
        # todo: add method
        pass



# Extract images from a PDF
from typing import Optional

from PIL import Image

from pdfconduit.utils.info import Info


# Todo: Fix img_extract and develop tests
def img_extract(pdf: str, password: Optional[str] = None) -> None:
    # Read PDF file
    reader = Info(pdf, password).pdf

    # Number of pages in input document
    page_count = reader.getNumPages()

    # 5c. Go through all the input file pages to add a watermark to them
    for page_number in range(page_count):
        # Merge the watermark with the page
        page = reader.getPage(page_number)

        xobj = page["/Resources"]["/XObject"].getObject()
        for obj in xobj:
            if xobj[obj]["/Subtype"] == "/Image":
                size = (xobj[obj]["/Width"], xobj[obj]["/Height"])
                data = xobj[obj].getData()
                if xobj[obj]["/ColorSpace"] == "/DeviceRGB":
                    mode = "RGB"
                else:
                    mode = "P"

                if xobj[obj]["/Filter"] == "/FlateDecode":
                    img = Image.frombytes(mode, size, data)
                    img.save(obj[1:] + ".png")  # TODO: Add save destination parameter
                elif xobj[obj]["/Filter"] == "/DCTDecode":
                    img = open(obj[1:] + ".jpg", "wb")
                    img.write(data)
                    img.close()
                elif xobj[obj]["/Filter"] == "/JPXDecode":
                    img = open(obj[1:] + ".jp2", "wb")
                    img.write(data)
                    img.close()


# TODO: Fix text extract and interpret extracted text
def text_extract(path, password=None):
    """Extract text from a PDF file"""
    pdf = Info(path, password).pdf

    return [pdf.getPage(i).extractText() for i in range(pdf.getNumPages())]



from dataclasses import dataclass
from enum import Enum
from typing import Optional

from pypdf.constants import UserAccessPermissions


class Algorithms(Enum):
    RC4_40: str = "RC4-40"
    RC4_128: str = "RC4-128"
    AES_128: str = "AES-128"
    AES_256: str = "AES-256"
    AES_256_r5: str = "AES-256-R5"

    @property
    def bit_length(self):
        if "40" in self.value:
            return 40
        elif "128" in self.value:
            return 128
        elif "256" in self.value:
            return 256

    @property
    def is_40bit(self) -> bool:
        return self.bit_length == 40

    @property
    def is_128bit(self) -> bool:
        return self.bit_length == 128

    @property
    def is_256bit(self) -> bool:
        return self.bit_length == 256

    @classmethod
    def from_algo(cls, algo: str):
        return cls(algo)


@dataclass
class Encryption:
    user_pw: str
    owner_pw: Optional[str]
    allow_printing: bool = True
    allow_commenting: bool = False
    algo: Algorithms = Algorithms.AES_256_r5
    permissions: UserAccessPermissions = UserAccessPermissions

    def __post_init__(self):
        if self.allow_printing and self.allow_commenting:
            self.permissions = (
                UserAccessPermissions.PRINT | UserAccessPermissions.MODIFY
            )
        elif self.allow_printing:
            self.permissions = UserAccessPermissions.PRINT
        elif self.allow_commenting:
            self.permissions = UserAccessPermissions.MODIFY
        else:
            self.permissions = UserAccessPermissions.R2



import json
import os


def main():
    from argparse import ArgumentParser

    # Declare argparse argument descriptions
    usage = "View test timings"
    description = "View test timings saved to json."

    # construct the argument parse and parse the arguments
    ap = ArgumentParser(usage=usage, description=description)
    ap.add_argument(
        "--times", action="store_true", help="Sort test results by execution time"
    )
    ap.add_argument(
        "--ascending", action="store_true", help="Order test results in ascending order"
    )
    args = ap.parse_args()

    json_path = os.path.join(os.path.dirname(__file__), "data", "timings.json")

    with open(json_path, "r") as json_file:
        data = json.load(json_file)

    if args.times:
        data = dict(
            sorted(
                data.items(),
                key=lambda item: item[1],
                reverse=args.ascending if args.ascending else False,
            )
        )

    for test_name, time in data.items():
        first = "{0:85}".format(test_name)
        print("{0:86} {1:7} msecs".format(first, time))


if __name__ == "__main__":
    main()



from pdfconduit.transform.merge import Merge
from pdfconduit.transform.merge2 import Merge2
from pdfconduit.transform.rotate import Rotate
from pdfconduit.transform.upscale import Upscale

__all__ = [Merge, Merge2, Rotate, Upscale]



from pdfconduit.convert.flatten import Flatten
from pdfconduit.convert.img2pdf import IMG2PDF
from pdfconduit.convert.pdf2img import PDF2IMG

# TODO: Create conversion modules pdf2docx, docx2pdf, html2pdf, csv2pdf, excel2pdf
__all__ = ["IMG2PDF", "PDF2IMG", "Flatten"]



# Convert a PNG image file to a PDF
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, List, Union

from PIL import Image

from pdfconduit.transform.merge import Merge
from pdfconduit.watermark.modify.canvas import CanvasImg, CanvasObjects
from pdfconduit.watermark.modify.draw import WatermarkDraw


class IMG2PDF:
    def __init__(
        self,
        imgs: Optional[List[str]] = None,
        destination: Optional[str] = None,
        tempdir: Optional[Union[str, TemporaryDirectory]] = None,
    ):
        """Convert each image into a PDF page and merge all pages to one PDF file"""
        self.imgs = imgs
        self.output_dir = destination
        if not tempdir:
            self._temp = TemporaryDirectory()
            self.tempdir = self._temp.name
        elif isinstance(tempdir, TemporaryDirectory):
            self._temp = tempdir
            self.tempdir = self._temp.name
        else:
            self.tempdir = tempdir

        self._pdf_pages = None

    @property
    def pdf_pages(self) -> List[str]:
        if not self._pdf_pages:
            self._pdf_pages = self.img2pdf()
        return self._pdf_pages

    def cleanup(self, clean_temp: bool = True) -> None:
        if clean_temp and hasattr(self, "_temp"):
            self._temp.cleanup()

    def _image_loop(self) -> List[str]:
        """Retrieve an iterable of images either with, or without a progress bar."""
        return self.imgs

    def _convert(self, image, output: Optional[str] = None) -> str:
        """Private method for converting a single PNG image to a PDF."""
        with Image.open(image) as im:
            width, height = im.size

            co = CanvasObjects()
            co.add(CanvasImg(image, 1.0, w=width, h=height, mask=None))

            return WatermarkDraw(
                co, tempdir=self.tempdir, pagesize=(width, height)
            ).write(output)

    def convert(self, image: str, output: Optional[str] = None):
        """
        Convert an image to a PDF.

        :param image: Image file path
        :param output: Output name, same as image name with .pdf extension by default
        :return: PDF file path
        """
        return self._convert(
            image, image.replace(Path(image).suffix, ".pdf") if not output else output
        )

    def img2pdf(self) -> List[str]:
        """Convert a list of images into a PDF files."""
        return [self._convert(image) for image in self._image_loop()]

    def save(self, output_name: str = "merged imgs", clean_temp: bool = True) -> str:
        m = Merge(self.pdf_pages, output_name=output_name, output_dir=self.output_dir)
        self.cleanup(clean_temp)
        return m.merge()



from pdfconduit.settings.compression import Compression, ImageQualityRange
from pdfconduit.settings.encryption import Algorithms, Encryption


__all__ = [Compression, ImageQualityRange, Encryption, Algorithms]



from typing import Optional

from pdfconduit.utils import Info, add_suffix
from pdfconduit.watermark.modify.canvas import CanvasObjects, CanvasStr
from pdfconduit.watermark.modify.draw import WatermarkDraw
from pdfconduit.watermark.watermark import Watermark


def mean(numbers):
    return float(sum(numbers)) / max(len(numbers), 1)


class Label(WatermarkDraw):
    def __init__(
        self,
        document: str,
        label: str,
        title_page: bool = False,
        suffix: Optional[str] = "labeled",
        output: Optional[str] = None,
        tempdir: Optional[str] = None,
    ):
        self.size = Info(document).size
        super(Label, self).__init__(
            self._create_canvas_objects(label, title_page),
            pagesize=self.size,
            tempdir=tempdir,
        )
        self.document = document
        self.watermark = self._write()

        if output:
            self.output = output
        elif suffix:
            suffix = label if not suffix else suffix
            self.output = add_suffix(self.document, suffix)
        else:
            self.output = self.dst

    def _create_canvas_objects(
        self, label: str, title_page: bool = False
    ) -> CanvasObjects:
        objects = CanvasObjects()
        if not title_page:
            objects.add(
                CanvasStr(
                    label,
                    size=int(mean(self.size) * 0.02),
                    opacity=1,
                    x=-(self.size[0] / 2) + 15,
                    y=-(self.size[1] / 2) + 25,
                    x_centered=False,
                )
            )
        else:
            objects.add(
                CanvasStr(
                    label, size=int(mean(self.size) * 0.1), opacity=1, y_centered=True
                )
            )
        return objects

    def write(self, cleanup: bool = True) -> str:
        wm = Watermark(
            self.document,
            tempdir=self.dir,
            use_receipt=False,
            remove_temps=True,
        )
        labeled = wm.add(watermark=self.watermark, output=self.output)
        if cleanup:
            wm.cleanup()
        return labeled



# Rotate a pdf file
import os
from io import BytesIO
from tempfile import NamedTemporaryFile
from typing import Optional, Union

from pdfrw import (
    PdfReader as PdfrwReader,
    PdfWriter as PdfrwWriter,
)
from pypdf import PdfReader as PypdfReader, PdfWriter as PypdfWriter

from pdfconduit.utils.driver import PdfDriver
from pdfconduit.utils.path import add_suffix


class Rotate(PdfDriver):

    def __init__(
        self,
        pdf: Union[str, BytesIO],
        rotation: int,
        suffix: str = "rotated",
        tempdir: Optional[str] = None,
        output: Optional[str] = None,
    ):
        self.pdf_object = pdf
        self.rotation = rotation
        self.suffix = suffix

        if output:
            self.outfn = output
        else:
            self.tempdir = tempdir

            if tempdir:
                with NamedTemporaryFile(
                    suffix=".pdf", dir=tempdir, delete=False
                ) as temp:
                    self.outfn = temp.name
            elif suffix:
                self.outfn = os.path.join(os.path.dirname(pdf), add_suffix(pdf, suffix))
            else:
                self.outfn = NamedTemporaryFile(suffix=".pdf").name

    def __str__(self) -> str:
        return self.file

    def rotate(self) -> str:
        return self.execute()

    @property
    def file(self) -> str:
        return str(self.outfn)

    def pdfrw(self) -> str:
        if isinstance(self.pdf_object, BytesIO):
            trailer = PdfrwReader(fdata=self.pdf_object.getvalue())
        else:
            trailer = PdfrwReader(fname=self.pdf_object)

        pages = trailer.pages

        ranges = [[1, len(pages)]]

        for onerange in ranges:
            onerange = (onerange + onerange[-1:])[:2]
            for pagenum in range(onerange[0] - 1, onerange[1]):
                pages[pagenum].Rotate = (
                    int(pages[pagenum].inheritable.Rotate or 0) + self.rotation
                ) % 360

        outdata = PdfrwWriter(self.outfn)
        outdata.trailer = trailer
        outdata.write()
        return self.outfn

    def pypdf(self) -> str:
        reader = PypdfReader(self.pdf_object)
        writer = PypdfWriter()

        for page_num in range(1, reader.get_num_pages()):
            writer.add_page(reader.pages[page_num]).rotate(self.rotation)

        with open(self.outfn, "wb") as fp:
            writer.write(fp)

        return self.outfn



from dataclasses import dataclass
from enum import Enum

from pdfconduit.utils.typing import List, Self


class Compression(Enum):
    DEFAULT: int = -1
    NONE: int = 0
    BEST_SPEED: int = 1
    BEST_COMPRESSION: int = 9
    LEVEL_0: int = 0
    LEVEL_1: int = 1
    LEVEL_2: int = 2
    LEVEL_3: int = 3
    LEVEL_4: int = 4
    LEVEL_5: int = 5
    LEVEL_6: int = 6
    LEVEL_7: int = 7
    LEVEL_8: int = 8
    LEVEL_9: int = 9

    @classmethod
    def from_level(cls, level):
        return cls(level)

    @classmethod
    def all(cls) -> List[Self]:
        return list(map(lambda c: c, cls))


@dataclass
class ImageQualityRange:
    min: int = 1
    max: int = 99



import os
import unittest
from tempfile import TemporaryDirectory
from typing import Tuple, List

from parameterized import parameterized

from pdfconduit import Info, Label, Watermark
from tests import *


def watermark_params() -> List[Tuple[str, str, bool, bool]]:
    return [
        # name, method, flatten, underneath
        ("basic", "pdfrw", False, False),
        ("underneath", "pdfrw", False, True),
        ("overlay", "pdfrw", False, False),
        ("flattened", "pdfrw", True, False),
        ("flattened_underneath", "pdfrw", True, True),
        ("basic", "pypdf", False, False),
        ("underneath", "pypdf", False, True),
        ("overlay", "pypdf", False, False),
        ("flattened", "pypdf", True, False),
        ("flattened_underneath", "pypdf", True, True),
    ]


def encryption_name_func(testcase_func, param_num, param):
    return "{}.{}.{}".format(testcase_func.__name__, param.args[0], param.args[1])


class TestWatermark(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pdf_path = pdf_path

    def setUp(self):
        self.temp = TemporaryDirectory()
        self.address = "43 Indian Lane"
        self.town = "Franklin"
        self.state = "MA"
        self.rotate = 30
        self.owner_pw = "foo"
        self.user_pw = "baz"

    def tearDown(self):
        self.temp.cleanup()

    @parameterized.expand(watermark_params, name_func=encryption_name_func)
    def test_watermark(
        self, name: str, method: str, flatten: bool = False, underneath: bool = False
    ):
        watermarker = Watermark(self.pdf_path, use_receipt=True, tempdir=self.temp.name)
        watermark = watermarker.draw(
            text1=self.address,
            text2=str(self.town + ", " + self.state),
            opacity=0.08,
            rotate=self.rotate,
            flatten=flatten,
        )
        added = watermarker.add(
            self.pdf_path, watermark, method=method, underneath=underneath
        )

        self.assertPdfExists(watermark)
        self.assertPdfExists(added)
        self.assertPdfHasResources(added)

    def test_conduit_watermark_label(self):
        """Apply a watermark label to a PDF file."""
        label = os.path.basename(self.pdf_path)
        labeled = Label(
            self.pdf_path, label, tempdir=self.temp.name, suffix=None
        ).write(cleanup=False)

        self.assertPdfExists(labeled)
        self.assertPdfHasResources(labeled)

    def assertPdfExists(self, pdf):
        # Assert watermarked PDF file exists
        self.assertTrue(os.path.exists(pdf))

    def assertPdfHasResources(self, pdf):
        # Assert watermarked PDF has page resources
        self.assertTrue(Info(pdf).resources())


if __name__ == "__main__":
    unittest.main()



# Merge PDF documents
import os
from typing import Union, Optional, List

from pdfrw import (
    PdfReader as PdfrwReader,
    PdfWriter as PdfrwWriter,
    IndirectPdfDict as PdfrwIndirectPdfDict,
)
from pypdf import PdfWriter as PyPdfWriter

from pdfconduit.utils.driver import PdfDriver

INPUT_PDFS_TYPE = Union[list, str]


class Merge(PdfDriver):
    file: str = None

    def __init__(
        self,
        input_pdfs: INPUT_PDFS_TYPE,
        output_name: str = "merged",
        output_dir: Optional[str] = None,
    ):
        self.pdfs = self._get_pdf_list(input_pdfs)
        self.directory = output_dir if output_dir else os.path.dirname(self.pdfs[0])
        self.output = os.path.join(
            self.directory, output_name.replace(".pdf", "") + ".pdf"
        )

    def __str__(self) -> str:
        return str(self.file)

    @staticmethod
    def validate(pdf) -> bool:
        # todo: simplify logic?
        if not pdf.startswith(".") and pdf.endswith(".pdf"):
            return True

    def _get_pdf_list(self, input_pdfs: INPUT_PDFS_TYPE) -> List[str]:
        """
        Generate list of PDF documents.

        :param input_pdfs: List of PDFs or a directory path
             Directory - Scans directory contents
             List - Filters list to assert all list items are paths to PDF documents
        :return: List of PDF paths
        """
        if isinstance(input_pdfs, list):
            return [pdf for pdf in input_pdfs if self.validate(pdf)]
        elif os.path.isdir(input_pdfs):
            return [
                os.path.join(input_pdfs, pdf)
                for pdf in os.listdir(input_pdfs)
                if self.validate(pdf)
            ]
        # todo: raise error if conditions aren't met?

    def merge(self) -> str:
        """Merge list of PDF files to a single PDF file."""
        self.file = self.execute()
        return self.file

    def pdfrw(self):
        writer = PdfrwWriter()
        for inpfn in self.pdfs:
            writer.addpages(PdfrwReader(inpfn).pages)

        writer.trailer.Info = PdfrwIndirectPdfDict(
            Author="Stephen Neal",
            Creator="pdfconduit",
            Producer="pdfconduit",
        )
        writer.write(self.output)
        return self.output

    def pypdf(self):
        merger = PyPdfWriter()

        for pdf in self.pdfs:
            merger.append(pdf)

        merger.write(self.output)
        merger.close()

        return self.output



from pdfconduit.watermark.modify.canvas.constructor import CanvasConstructor
from pdfconduit.watermark.modify.canvas.objects import (
    CanvasImg,
    CanvasStr,
    CanvasObjects,
)

__all__ = ["CanvasConstructor", "CanvasImg", "CanvasStr", "CanvasObjects"]



import os.path

from parameterized import parameterized

from pdfconduit import Pdfconduit, Info
from pdfconduit.utils.typing import List, Iterable
from tests import PdfconduitTestCase, get_clean_pdf_name
from tests import test_data_path


def merge_params():
    params = [
        ("document.pdf", ["article.pdf", "manual.pdf"]),
        ("document.pdf", ["workbook.pdf", "article.pdf"]),
        ("workbook.pdf", ["document.pdf", "article.pdf"]),
        ("article.pdf", ["document.pdf"]),
        ("document.pdf", ["workbook.pdf", "document.pdf"]),
    ]
    return [
        (test_data_path(main_pdf), list(map(lambda pdf: test_data_path(pdf), to_merge)))
        for main_pdf, to_merge in params
    ]


def merge_name_func(testcase_func, param_num, param):
    return "{}.{}_{}".format(
        testcase_func.__name__,
        param_num,
        get_clean_pdf_name(param.args[0]),
    )


class TestMerge(PdfconduitTestCase):
    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_pdfs(self, main_pdf: str, pdfs_to_merge: Iterable[str]):
        self.pdf_path = main_pdf
        self.conduit = Pdfconduit(main_pdf).set_output_directory(self.temp.name)

        for pdf in pdfs_to_merge:
            self.conduit.merge(pdf)

        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertCorrectNumPages(main_pdf, pdfs_to_merge, self.conduit.info.pages)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_pdfs_using_context(
        self, main_pdf: str, pdfs_to_merge: Iterable[str]
    ):
        self.pdf_path = main_pdf
        with Pdfconduit(self.pdf_path) as conduit:
            conduit.set_output_directory(self.temp.name)
            for pdf in pdfs_to_merge:
                conduit.merge(pdf)

        self.assertPdfExists(conduit.output)
        self.assertCorrectNumPages(main_pdf, pdfs_to_merge, conduit.info.pages)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_pdfs_fast(self, main_pdf: str, pdfs_to_merge: List[str]):
        self.pdf_path = main_pdf
        self.conduit = Pdfconduit(main_pdf).set_output_directory(self.temp.name)

        self.conduit.merge_fast(pdfs_to_merge)

        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertCorrectNumPages(main_pdf, pdfs_to_merge, self.conduit.info.pages)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_pdfs_fast_from_stream(
        self, main_pdf: str, pdfs_to_merge: List[str]
    ):
        self.pdf_path = main_pdf
        self.conduit = Pdfconduit(
            self._get_pdf_byte_stream(self.pdf_path)
        ).set_output_directory(self.temp.name)

        self.conduit.merge_fast(pdfs_to_merge)

        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertCorrectNumPages(main_pdf, pdfs_to_merge, self.conduit.info.pages)



# Merge PDF documents
from io import BytesIO
from typing import Union, Iterable

from pdfrw import (
    PdfReader as PdfrwReader,
    PdfWriter as PdfrwWriter,
    IndirectPdfDict as PdfrwIndirectPdfDict,
)
from pypdf import PdfWriter as PyPdfWriter

from pdfconduit.utils.driver import PdfDriver


class Merge2(PdfDriver):
    def __init__(self, pdfs: Iterable[Union[str, BytesIO]], output: str):
        self._pdfs = pdfs
        self._output = output

    def merge(self) -> str:
        """Merge list of PDF files to a single PDF file."""
        return self.execute()

    def pdfrw(self):
        writer = PdfrwWriter()

        for pdf_object in self._pdfs:
            if isinstance(pdf_object, BytesIO):
                reader = PdfrwReader(fdata=pdf_object.getvalue())
            else:
                reader = PdfrwReader(fname=pdf_object)
            writer.addpages(reader.pages)

        writer.trailer.Info = PdfrwIndirectPdfDict(
            Author="Stephen Neal",
            Creator="pdfconduit",
            Producer="pdfconduit",
        )
        writer.write(self._output)
        return self._output

    def pypdf(self):
        merger = PyPdfWriter()

        for pdf in self._pdfs:
            merger.append(pdf)

        merger.write(self._output)
        merger.close()

        return self._output



# Upscale a PDF file
import os
from tempfile import NamedTemporaryFile
from typing import Optional

from pdfrw import (
    PdfReader as pdfrwReader,
    PdfWriter as pdfrwWriter,
    PageMerge as pdfrwPageMerge,
    IndirectPdfDict as pdfrwIndirectPdfDict,
)
from pypdf import (
    PdfReader as pypdfReader,
    PdfWriter as pypdfWriter,
)

from pdfconduit.utils.driver import PdfDriver
from pdfconduit.utils.info import Info
from pdfconduit.utils.path import add_suffix


class Upscale(PdfDriver):
    def __init__(
        self,
        file_name: str,
        margin_x: int = 0,
        margin_y: int = 0,
        scale: float = 1.5,
        suffix: str = "scaled",
        tempdir: Optional[str] = None,
    ):
        self.file_name = file_name
        self.margin_x = margin_x
        self.margin_y = margin_y
        self.scale = scale
        self.suffix = suffix

        # Set output file name
        if tempdir:
            with NamedTemporaryFile(
                suffix="_" + suffix + ".pdf", dir=tempdir, delete=False
            ) as temp:
                self.output = temp.name
        elif suffix:
            self.output = os.path.join(
                os.path.dirname(file_name), add_suffix(file_name, suffix)
            )
        else:
            self.output = NamedTemporaryFile(suffix="_" + suffix + ".pdf").name

        # Get target width and height
        dims = Info(self.file_name).dimensions
        self.target_w = dims["w"] * self.scale
        self.target_h = dims["h"] * self.scale

    def __str__(self) -> str:
        return self.file

    def upscale(self) -> str:
        # Execute either pdfrw or PyPDF3 method
        return self.execute()

    @property
    def file(self) -> str:
        return str(self.output)

    def _pdfrw_adjust(self, page: object):
        info = pdfrwPageMerge().add(page)
        x1, y1, x2, y2 = info.xobj_box
        viewrect = (
            self.margin_x,
            self.margin_y,
            x2 - x1 - 2 * self.margin_x,
            y2 - y1 - 2 * self.margin_y,
        )
        page = pdfrwPageMerge().add(page, viewrect=viewrect)
        page[0].scale(self.scale)
        return page.render()

    def pdfrw(self) -> str:
        reader = pdfrwReader(self.file_name)
        writer = pdfrwWriter(self.output)
        for i in list(range(0, len(reader.pages))):
            writer.addpage(self._pdfrw_adjust(reader.pages[i]))
        writer.trailer.Info = pdfrwIndirectPdfDict(reader.Info or {})
        writer.write()

        return self.output

    def pypdf(self) -> str:
        reader = pypdfReader(self.file_name)
        writer = pypdfWriter()

        for page_num in range(0, reader.get_num_pages()):
            page = reader.pages[page_num]

            page.scale_to(width=self.target_w, height=self.target_h)

            writer.add_page(page)

        with open(self.output, "wb") as fp:
            writer.write(fp)

        return self.output



from parameterized import parameterized

from pdfconduit import Pdfconduit, Info
from pdfconduit.settings import Encryption, Compression
from pdfconduit.utils.typing import List, Iterable
from tests import *
from tests.test_encrypt import EncryptionTestCase
from tests.test_merge import merge_name_func


def stack_params() -> List[str]:
    return list(
        map(
            lambda filename: test_data_path(filename),
            [
                "article.pdf",
                "charts.pdf",
                "document.pdf",
                # "workbook.pdf",
            ],
        )
    )


def stack_name_func(testcase_func, param_num, param):
    return "{}.{}".format(
        testcase_func.__name__,
        get_clean_pdf_name(param.args[0]),
    )


def merge_params():
    params = [
        ("document.pdf", ["article.pdf", "charts.pdf"]),
        ("article.pdf", ["document.pdf"]),
        ("charts.pdf", ["article.pdf", "document.pdf"]),
    ]
    return [
        (test_data_path(main_pdf), list(map(lambda pdf: test_data_path(pdf), to_merge)))
        for main_pdf, to_merge in params
    ]


ENCRYPTOR = Encryption(user_pw="baz", owner_pw="foo")


class TestStacks(EncryptionTestCase):
    def createConduit(self, path: str, suffix: str):
        return (
            Pdfconduit(path)
            .set_output_directory(self.temp.name)
            .set_output_suffix(suffix)
        )

    def assertEncryption(self, conduit: Pdfconduit):
        security = self._getPdfSecurity(conduit)

        self.assertEncrypted(conduit)
        self.assert256BitEncryption(security, 5)
        self.assertPermissions(
            conduit,
            can_print=ENCRYPTOR.allow_printing,
            can_modify=ENCRYPTOR.allow_commenting,
        )

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_and_encrypt(self, path: str, pdfs_to_merge: Iterable[str]):
        self.conduit = self.createConduit(path, "merged_encrypted")
        for pdf in pdfs_to_merge:
            self.conduit.merge(pdf)

        self.assertCorrectNumPages(path, pdfs_to_merge, self.conduit.info.pages)

        self.conduit.encrypt(ENCRYPTOR)
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertEncryption(self.conduit)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_fast_and_encrypt(self, path: str, pdfs_to_merge: Iterable[str]):
        self.conduit = self.createConduit(path, "merged_fast_encrypted")
        self.conduit.merge_fast(list(pdfs_to_merge))

        self.assertCorrectNumPages(path, pdfs_to_merge, self.conduit.info.pages)

        self.conduit.encrypt(ENCRYPTOR)
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertEncryption(self.conduit)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_and_rotate(self, path: str, pdfs_to_merge: Iterable[str]):
        self.conduit = self.createConduit(path, "merged_rotated")
        for pdf in pdfs_to_merge:
            self.conduit.merge(pdf)
        self.conduit.rotate(90)
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertCorrectNumPages(path, pdfs_to_merge, self.conduit.info.pages)
        self.assertPdfRotation(self.conduit, 90)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_rotate_and_encrypt(
        self, path: str, pdfs_to_merge: Iterable[str]
    ):
        self.conduit = self.createConduit(path, "merged_rotated_encrypted")
        for pdf in pdfs_to_merge:
            self.conduit.merge(pdf)

        self.assertCorrectNumPages(path, pdfs_to_merge, self.conduit.info.pages)

        self.conduit.rotate(90)

        self.assertPdfRotation(self.conduit, 90)

        self.conduit.encrypt(ENCRYPTOR)
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertEncryption(self.conduit)

    @parameterized.expand(stack_params, name_func=stack_name_func)
    def test_can_rotate_and_encrypt(self, path: str):
        self.conduit = self.createConduit(path, "rotated_encrypted")
        self.conduit.rotate(90)

        self.assertPdfRotation(self.conduit, 90)

        self.conduit.encrypt(ENCRYPTOR)
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertEncryption(self.conduit)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_and_slice(self, path: str, pdfs_to_merge: Iterable[str]):
        self.conduit = self.createConduit(path, "merged_sliced")
        for pdf in pdfs_to_merge:
            self.conduit.merge(pdf)
        self.conduit.slice(2, 5)
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertCorrectPagesSliced(2, 5, self.conduit)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_slice_and_encrypt(self, path: str, pdfs_to_merge: Iterable[str]):
        self.conduit = self.createConduit(path, "merged_sliced_encrypted")
        for pdf in pdfs_to_merge:
            self.conduit.merge(pdf)

        self.assertCorrectNumPages(path, pdfs_to_merge, self.conduit.info.pages)

        self.conduit.slice(2, 5)

        self.assertCorrectPagesSliced(2, 5, self.conduit)

        self.conduit.encrypt(ENCRYPTOR)
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertEncryption(self.conduit)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_and_scale(self, path: str, pdfs_to_merge: Iterable[str]):
        self.conduit = self.createConduit(path, "merged_scaled")
        for pdf in pdfs_to_merge:
            self.conduit.merge(pdf)
        self.conduit.scale(2.0)
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertCorrectNumPages(path, pdfs_to_merge, self.conduit.info.pages)
        self.assertPdfScaled(2.0, self.conduit, path)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_scale_and_encrypt(self, path: str, pdfs_to_merge: Iterable[str]):
        self.conduit = self.createConduit(path, "merged_scaled_encrypted")
        for pdf in pdfs_to_merge:
            self.conduit.merge(pdf)

        self.assertCorrectNumPages(path, pdfs_to_merge, self.conduit.info.pages)

        self.conduit.scale(2.0)

        self.assertPdfScaled(2.0, self.conduit, path)

        self.conduit.encrypt(ENCRYPTOR)
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertEncryption(self.conduit)

    @parameterized.expand(stack_params, name_func=stack_name_func)
    def test_can_scale_and_flatten(self, path: str):
        self.conduit = self.createConduit(path, "scaled_flattened")
        self.conduit.scale(2.0)
        self.conduit.flatten()

        self.assertPdfExists(self.conduit.output)
        self.assertPdfScaled(2.0, self.conduit, path)
        self.assertPdfPagesEqual(path, self.conduit.output)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_and_flatten(self, path: str, pdfs_to_merge: Iterable[str]):
        self.conduit = self.createConduit(path, "merged_flattened")
        for pdf in pdfs_to_merge:
            self.conduit.merge(pdf)
        self.conduit.flatten()

        self.assertPdfExists(self.conduit.output)
        self.assertCorrectNumPages(path, pdfs_to_merge, self.conduit.info.pages)

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_and_remove_duplication(
        self, path: str, pdfs_to_merge: Iterable[str]
    ):
        self.conduit = self.createConduit(path, "merged_noimages")
        for pdf in pdfs_to_merge:
            self.conduit.merge(pdf)
        self.conduit.remove_images()
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertEqual(0, self.conduit.info.images_count)

        self.assertCorrectNumPages(path, pdfs_to_merge, self.conduit.info.pages)

    @parameterized.expand(stack_params, name_func=stack_name_func)
    def test_can_scale_and_compress(self, path: str):
        self.conduit = self.createConduit(path, "scaled_compressed")
        self.conduit.scale(2.0)
        self.conduit.compress()
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertPdfScaled(2.0, self.conduit, path)
        self.assertEqual(Info(path).images_count, self.conduit.info.images_count)

    @parameterized.expand(stack_params, name_func=stack_name_func)
    def test_can_scale_and_reduce_image_quality(self, path: str):
        self.conduit = self.createConduit(path, "scaled_reduced_image_quality")
        self.conduit.scale(2.0)
        self.conduit.reduce_image_quality(50)
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertEqual(Info(path).images_count, self.conduit.info.images_count)
        if path.endswith("document.pdf"):
            print("Skipping document.pdf because it yields the same size")
            return
        self.assertFileSizeDecreased(
            self.createConduit(path, "scaled").scale(2.0).write(), self.conduit.output
        )

    @parameterized.expand(merge_params, name_func=merge_name_func)
    def test_can_merge_scale_flatten_and_encrypt(
        self, path: str, pdfs_to_merge: Iterable[str]
    ):
        self.conduit = self.createConduit(path, "merged_scaled_flattened_encrypted")
        for pdf in pdfs_to_merge:
            self.conduit.merge(pdf)

        self.assertCorrectNumPages(path, pdfs_to_merge, self.conduit.info.pages)

        self.conduit.scale(2.0)

        self.assertPdfScaled(2.0, self.conduit, path)

        self.conduit.flatten()

        self.assertPdfExists(self.conduit.output)
        self.assertCorrectNumPages(path, pdfs_to_merge, self.conduit.info.pages)

        self.conduit.encrypt(ENCRYPTOR)
        self.conduit.write()

        self.assertPdfExists(self.conduit.output)
        self.assertEncryption(self.conduit)



from typing import List

from parameterized import parameterized

from pdfconduit import Info
from tests import PdfconduitTestCase


def slice_params() -> List[str]:
    return [
        "1to1",
        "4to7",
        "2to6",
        "1to8",
        "4to9",
    ]


def slice_name_func(testcase_func, param_num, param):
    return "{}.{}".format(
        testcase_func.__name__,
        param.args[0],
    )


class TestSlice(PdfconduitTestCase):
    @parameterized.expand(slice_params, name_func=slice_name_func)
    def test_slice(self, slice_range: str):
        fp, lp = tuple(slice_range.split("to"))
        fp = int(fp)
        lp = int(lp)

        self.conduit.set_output_suffix("sliced_{}".format(slice_range)).slice(
            fp, lp
        ).write()

        self.assertPdfExists(self.conduit.output)
        self.assertCorrectPagesSliced(fp, lp, self.conduit)



# todo: cleanup imports
from pdfconduit.pdfconduit import Pdfconduit
from pdfconduit.transform import Merge, Rotate, Upscale
from pdfconduit.utils import Info
from pdfconduit.watermark.label import Label
from pdfconduit.watermark.watermark import Watermark

__all__ = [Merge, Rotate, Upscale, Info, Pdfconduit, Watermark, Label]



from typing import Dict

from pypdf import PdfReader


class Permissions:
    def __init__(self, pdf: PdfReader):
        self._permissions = pdf.user_access_permissions.to_dict()

    def all(self) -> Dict[str, bool]:
        return self._permissions

    def can(self, permission: str) -> bool:
        return permission in self._permissions and self._permissions[permission]

    def can_print(self) -> bool:
        return self.can("print")

    def can_modify(self) -> bool:
        return self.can("modify")

    def can_copy(self) -> bool:
        return self.can("copy")

    def can_annotate(self) -> bool:
        return self.can("annotations")

    def can_fill_forms(self) -> bool:
        return self.can("forms")

    def can_change_accessability(self) -> bool:
        return self.can("accessability")

    def can_assemble(self) -> bool:
        return self.can("assemble")

    def can_print_high_quality(self) -> bool:
        return self.can("print_high_quality")



import random
from typing import List

from parameterized import parameterized

from pdfconduit import Info, Pdfconduit
from tests import PdfconduitTestCase


def rotate_params() -> List[int]:
    return [
        90,
        180,
        270,
    ]


def rotate_exact_params() -> List[int]:
    return list(
        map(lambda rotation: rotation + random.randrange(10, 80), rotate_params())
    )


def rotate_name_func(testcase_func, param_num, param):
    return "{}.{}".format(
        testcase_func.__name__,
        param.args[0],
    )


class TestRotate(PdfconduitTestCase):
    @parameterized.expand(rotate_params, name_func=rotate_name_func)
    def test_can_rotate(self, rotation: int):
        self.conduit.rotate(rotation).set_output_suffix(
            "rotated_{}".format(rotation)
        ).write()

        self.assertPdfExists(self.conduit.output)
        self.assertPdfRotation(self.conduit, rotation)

    @parameterized.expand(rotate_exact_params, name_func=rotate_name_func)
    def test_can_rotate_exact(self, rotation: int):
        self.conduit.rotate_exact(rotation).set_output_suffix(
            "rotated_{}".format(rotation)
        ).write()

        self.assertPdfExists(self.conduit.output)
        self.assertPdfRotation(self.conduit, rotation)

    @parameterized.expand(rotate_exact_params, name_func=rotate_name_func)
    def test_cannot_rotate(self, rotation: int):
        with self.assertRaises(ValueError) as context:
            self.conduit.rotate(rotation).write()

        self.assertPdfDoesntExists(self.conduit.output)
        self.assertTrue(
            "Rotation angle must be a multiple of 90" in str(context.exception)
        )

    @parameterized.expand(rotate_exact_params, name_func=rotate_name_func)
    def test_can_rotate_exact_from_stream(self, rotation: int):
        self.conduit = Pdfconduit(
            self._get_pdf_byte_stream(self.pdf_path)
        ).set_output_directory(self.temp.name)
        self.conduit.rotate_exact(rotation).set_output_suffix(
            "rotated_{}".format(rotation)
        ).write()

        self.assertPdfExists(self.conduit.output)
        self.assertPdfRotation(self.conduit, rotation)



from abc import ABC, abstractmethod
from enum import Enum

from pdfconduit.utils.typing import Self


class Driver(Enum):
    pdfrw: str = "pdfrw"
    pypdf: str = "pypdf"


class PdfDriver(ABC):
    _driver: Driver = Driver.pdfrw

    def use_pdfrw(self) -> Self:
        return self.use(Driver.pdfrw)

    def use_pypdf(self) -> Self:
        return self.use(Driver.pypdf)

    def use(self, driver: Driver) -> Self:
        self._driver = driver
        return self

    def is_driver_pdfrw(self) -> bool:
        return self._driver == Driver.pdfrw

    def is_driver_pypdf(self) -> bool:
        return self._driver == Driver.pypdf

    @abstractmethod
    def pdfrw(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def pypdf(self) -> str:
        raise NotImplementedError

    def execute(self) -> str:
        return self.pdfrw() if self.is_driver_pdfrw() else self.pypdf()



from typing import Tuple, List

from parameterized import parameterized

from pdfconduit import Info
from tests import PdfconduitTestCase


def scale_params() -> List[Tuple[float, bool]]:
    return [(scale, accel) for scale in [0.5, 1.5, 2.0, 3.0] for accel in [True, False]]


def scale_name_func(testcase_func, param_num, param):
    name = "{}.{}x".format(testcase_func.__name__, str(param.args[0]).replace(".", ","))
    return name + "_accelerate" if param.args[1] else name


class TestScale(PdfconduitTestCase):
    @parameterized.expand(scale_params, name_func=scale_name_func)
    def test_scale(self, scale: float, accelerate: bool):
        suffix = "scaled_{}x".format(scale)
        if accelerate:
            suffix += "_accelerated"

        self.conduit.scale(scale, accelerate=accelerate).set_output_suffix(
            suffix
        ).write()

        self.assertPdfExists(self.conduit.output)
        self.assertPdfScaled(scale, self.conduit.output)

    def assertPdfScaled(self, scale, scaled):
        self.assertEqual(
            Info(scaled).size, tuple([i * scale for i in Info(self.pdf_path).size])
        )
        self.assertEqual(Info(scaled).pages, Info(self.pdf_path).pages)



# Apply a watermark to a PDF file
import os
import shutil
from tempfile import TemporaryDirectory
from typing import Optional

from looptools import Timer

from pdfconduit.pdfconduit import Pdfconduit
from pdfconduit.settings import Encryption
from pdfconduit.utils import add_suffix, Info
from pdfconduit.watermark.add import WatermarkAdd
from pdfconduit.watermark.lib import IMAGE_DEFAULT, IMAGE_DIRECTORY
from pdfconduit.watermark.modify.canvas import CanvasConstructor
from pdfconduit.watermark.modify.draw import WatermarkDraw


class Watermark:
    def __init__(
        self,
        document: str,
        remove_temps: bool = True,
        move_temps: Optional[bool] = None,
        tempdir: Optional[str] = None,
        use_receipt: bool = True,
    ):
        """
        Watermark and encrypt a PDF document.

        Manage watermarking processes from single class initialization.  This class utilizes the draw,
        add and encrypt modules.

        :param document: str
            PDF document full path
        :param remove_temps: bool
            Remove temporary files after completion
        :param open_file: bool
            Open file after completion
        :param tempdir: function or str
            Temporary directory for file writing
        :param receipt: cls
            Use existing Receipt object if already initiated
        :param use_receipt: bool
            Print receipt information to console and write to file
        """
        self.time = Timer()
        self.document_og = document
        self.document = self.document_og
        self.watermark = None
        self.remove_temps = remove_temps
        self.move_temps = move_temps

        if not tempdir:
            self._temp = TemporaryDirectory()
            self.tempdir = self._temp.name
        elif isinstance(tempdir, TemporaryDirectory):
            self._temp = tempdir
            self.tempdir = self._temp.name
        else:
            self.tempdir = tempdir

    def __str__(self) -> str:
        return str(self.document)

    def cleanup(self) -> str:
        runtime = self.time.end
        if self.move_temps:
            if os.path.isdir(self.move_temps):
                shutil.move(self.tempdir, self.move_temps)
        if self.remove_temps:
            if os.path.isdir(self.tempdir):
                shutil.rmtree(self.tempdir)
        return self.document

    def draw(
        self,
        text1: Optional[str] = None,
        text2: Optional[str] = None,
        include_copyright: bool = True,
        image: str = IMAGE_DEFAULT,
        rotate: int = 30,
        opacity: float = 0.08,
        compress: int = 0,
        flatten: bool = False,
        add: bool = False,
    ) -> str:
        """
        Draw watermark PDF file.

        Create watermark using either a reportlabs canvas or a PIL image.

        :param text1: str
            Text line 1
        :param text2: str
            Text line 2
        :param include_copyright: bool
            Draw copyright and year to canvas
        :param image: str
            Logo image to be used as base watermark
        :param rotate: int
            Degrees to rotate canvas by
        :param opacity: float
            Watermark opacity
        :param compress: bool
            Compress watermark contents  (not entire PDF)
        :param flatten: bool
            Draw watermark with multiple layers or a single flattened layer
        :param add: bool
            Add watermark to original document
        :return: str
            Watermark PDF file full path
        """
        im_path = os.path.join(IMAGE_DIRECTORY, image)
        if os.path.isfile(im_path):
            image = im_path

        co = CanvasConstructor(
            text1,
            text2,
            include_copyright,
            image,
            rotate,
            opacity,
            tempdir=self.tempdir,
        )
        objects, rotate = (
            co.img() if flatten else co.canvas()
        )  # Run img constructor method if flatten is True

        # Draw watermark to file
        self.watermark = WatermarkDraw(
            objects,
            rotate=rotate,
            compress=compress,
            tempdir=self.tempdir,
            pagesize=Info(self.document_og).size,
            pagescale=True,
        ).write()

        if not add:
            return self.watermark
        else:
            self.add()
            return self.cleanup()

    def add(
        self,
        document: Optional[str] = None,
        watermark: Optional[str] = None,
        underneath: bool = False,
        output: Optional[str] = None,
        suffix: Optional[str] = "watermarked",
        method: str = "pdfrw",
    ) -> str:
        """
        Add a watermark file to an existing PDF document.

        Rotate and upscale watermark file as needed to fit existing PDF document.  Watermark can be overlayed or
        placed underneath.

        :param document: str
            PDF document full path
        :param watermark: str
            Watermark PDF full path
        :param underneath: bool
            Place watermark either under or over existing PDF document
        :param output: str
            Output file path
        :param suffix: str
            Suffix to append to existing PDF document file name
        :param method: str
            PDF library to be used for watermark adding
        :return: str
            Watermarked PDF Document full path
        """
        if not watermark:
            watermark = self.watermark
        if not document:
            document = self.document

        watermarker = WatermarkAdd(
            document,
            watermark,
            output=output,
            underneath=underneath,
            tempdir=self.tempdir,
            suffix=suffix,
        )
        if method == "pdfrw":
            watermarker.use_pdfrw()
        else:
            watermarker.use_pypdf()
        self.document = watermarker.add()

        return self.document

    def encrypt(
        self,
        user_pw: str = "",
        owner_pw: Optional[str] = None,
        encrypt_128: bool = True,
        allow_printing: bool = True,
        allow_commenting: bool = False,
        document: Optional[str] = None,
    ):
        """
        Encrypt a PDF document to add passwords and restrict permissions.

        Add a user password that must be entered to view document and a owner password that must be entered to alter
        permissions and security settings.  Encryption keys are 128 bit when encrypt_128 is True and 40 bit when
        False.  By default permissions are restricted to print only, when set to false all permissions are allowed.
        TODO: Add additional permission parameters

            User password required to open and view PDF document
        :param owner_pw: str
            Owner password required to alter security settings and permissions
        :param encrypt_128: bool
            Encrypt PDF document using 128 bit keys
        :param allow_printing: bool
            Restrict permissions to print only
        :param allow_commenting:
        :param user_pw: str
        :param document:
        :return: str
            Encrypted PDF full path
        """
        document = self.document if document is None else document

        encrypter = Encryption(
            user_pw=user_pw,
            owner_pw=owner_pw,
            allow_printing=allow_printing,
            allow_commenting=allow_commenting,
        )
        p = (
            Pdfconduit(document)
            .set_output(add_suffix(self.document_og, "secured"))
            .encrypt(encrypter)
            .write()
        )

        return p



import os
from setuptools import setup, find_packages

long_description = """
A Pure-Python library built as a PDF toolkit.  Prepare documents for distribution.

Features:
- Watermark: Dynamically generate watermarks and add watermark to existing document
- Label: Overlay text labels such as filename or date to documents 
- Encrypt: Password protect and restrict permissions to print only
- Rotate: Rotate by increments of 90 degrees
- Upscale: Scale PDF size
- Merge: Concatenate multiple documents into one file
- Slice: Extract page ranges from documents
- Extract Text and Images
- Retrieve document metadata and information
"""


def get_version(version_file='_version.py'):
    """Retrieve the package version from a version file in the package root."""
    with open(os.path.join(os.path.dirname(__file__), 'pdfconduit', version_file), 'rb') as fp:
        return fp.read().decode('utf8').split('=')[1].strip("\n").replace('"', '')


setup(
    name='pdfconduit',
    version=get_version(),
    packages=find_packages(),
    install_requires=[
        'looptools',
        'pdfrw>=0.4',
        'PyMuPDF>=1.17.7,<=1.24.9',
        'pypdf[crypto]>=4.2.0',
        'PillowImage>=1.2.0',
        'PyBundle>=1.0.6',
    ],
    python_requires='>3.6.8',
    include_package_data=True,
    url='https://github.com/sfneal/pdfconduit',
    license='',
    author='Stephen Neal',
    author_email='stephen@stephenneal.net',
    description='PDF toolkit for preparing documents for distribution.',
    long_description=long_description,
)



# Create flat PDF by converting each input PDF page to a PNG
import os
from tempfile import TemporaryDirectory
from typing import Optional, List

from pdfconduit.convert.img2pdf import IMG2PDF
from pdfconduit.convert.pdf2img import PDF2IMG
from pdfconduit.transform.upscale import Upscale
from pdfconduit.utils.path import add_suffix


class Flatten:
    def __init__(
        self,
        file_name: str,
        scale: float = 1.0,
        suffix: str = "flat",
        tempdir: Optional[str] = None,
    ):
        """Create a flat single-layer PDF by converting each page to a PNG image"""
        self._file_name = file_name

        if not tempdir:
            self._temp = TemporaryDirectory()
            self.tempdir = self._temp.name
        elif isinstance(tempdir, TemporaryDirectory):
            self._temp = tempdir
            self.tempdir = self._temp.name
        else:
            self.tempdir = tempdir

        self.suffix = suffix
        self.directory = os.path.dirname(file_name)

        if scale and scale != 0 and scale != 1.0:
            self.file_name = Upscale(
                file_name, scale=scale, tempdir=self.tempdir
            ).upscale()
        else:
            self.file_name = self._file_name

        self.imgs = None
        self.pdf = None

    def __str__(self) -> str:
        return str(self.pdf)

    def get_imgs(self) -> List[str]:
        self.imgs = PDF2IMG(self.file_name, tempdir=self.tempdir).save()
        return self.imgs

    def save(self, remove_temps: bool = True) -> str:
        if self.imgs is None:
            self.get_imgs()
        i2p = IMG2PDF(self.imgs, self.directory, self.tempdir)
        self.pdf = i2p.save(
            clean_temp=False, output_name=add_suffix(self._file_name, self.suffix)
        )
        self.cleanup(remove_temps)
        return self.pdf

    def cleanup(self, clean_temp: bool = True) -> None:
        if clean_temp and hasattr(self, "_temp"):
            self._temp.cleanup()



# Add a watermark PDF file to another PDF file
from tempfile import NamedTemporaryFile
from typing import Optional, Tuple

from PyBundle import resource_path
from pdfrw import PdfReader, PdfWriter, PageMerge
from pypdf import PdfReader as PypdfReader, PdfWriter as PypdfWriter
from reportlab.lib.pagesizes import letter

from pdfconduit.transform import Rotate, Upscale
from pdfconduit.utils import add_suffix, Info, pypdf_reader
from pdfconduit.utils.driver import PdfDriver


class WatermarkAdd(PdfDriver):
    def __init__(
        self,
        document: str,
        watermark: str,
        underneath: bool = False,
        overwrite: bool = False,
        output: Optional[str] = None,
        suffix: Optional[str] = "watermarked",
        decrypt: bool = False,
        tempdir: Optional[str] = None,
    ):
        """
        Add a watermark to an existing PDF document

        1. Read PDF document and decrypt if encrypted
        2. Get input PDF document info
            2a. Retrieve dimensions
            2b. Determine orientation
            2c. Upscale to fit letter
        3. Get watermark info
            3a. Retrieve dimensions
            3b. Rotate if necessary
            3c. Upscale to fit document
        4. Set file names
            4a. Check if a upscaled doc was created, if not use original document
            4b. Check if watermark was rotated, if not use original watermark
        5. Add watermark to PDF document
            5a. Create watermarked document filename (ex: originalpdfname_watermarked.pdf)
            5b. Create PDF file reader and file writer objects
            5c. Add watermark to each page of original
            5d. Save watermarked document to file
        """
        self.rotate = 0
        self.scale = 0
        self.underneath = underneath
        self.tempdir = tempdir

        self.document_reader = pypdf_reader(document, decrypt)
        self.document = self._get_document_info(document)
        self.watermark_file = self._get_watermark_info(self.document, watermark)
        self.pdf_fname, self.wtrmrk_fname = self._set_filenames

        if overwrite:
            self.output_filename = document
        elif output:
            self.output_filename = output
        elif suffix:
            self.output_filename = add_suffix(document, suffix)
        else:
            tmpf = NamedTemporaryFile(suffix=".pdf", dir=self.tempdir, delete=False)
            self.output_filename = resource_path(tmpf.name)

    def __str__(self):
        return str(self.output_filename)

    def _get_document_info(self, filename: str) -> dict:
        # todo: add use of typed dictionary
        pdf_file = {"path": filename}

        # 2a. Get PDF width and height
        # todo: optimize so that a pdf reader can be passed
        pdf_file.update(Info(filename).dimensions)

        # 2b. Get PDF file orientation
        if pdf_file["h"] > pdf_file["w"]:
            pdf_file["orientation"] = "portrait"
            letter_size = {"w": int(letter[0]), "h": int(letter[1])}
        else:
            pdf_file["orientation"] = "landscape"
            letter_size = {"h": int(letter[0]), "w": int(letter[1])}

        # 2c. Upscale PDF if it is smaller than a letter
        if pdf_file["w"] <= letter_size["w"] or pdf_file["h"] <= letter_size["h"]:
            scale = float(letter_size["w"] / pdf_file["w"])
            # todo: add use of upscale class instead of function
            pdf_file["upscaled"] = (
                Upscale(pdf_file["path"], scale=scale, tempdir=self.tempdir)
                .use(self._driver)
                .upscale()
            )
            self.document_reader = pypdf_reader(pdf_file["upscaled"])
        return pdf_file

    def _get_watermark_info(
        self, document: dict, watermark: str, margin_x: int = 0, margin_y: int = 0
    ) -> dict:
        # todo: add use of typed dictionary
        # 3a. Get watermark path and dimensions
        watermark_file = {"path": watermark}
        watermark_file.update(Info(watermark).dimensions)

        # 3b. Check if watermark file needs to be rotated
        if (
            watermark_file["w"] > watermark_file["h"]
            and document["orientation"] == "portrait"
        ):
            self.rotate = 90
            watermark_file["rotated"] = (
                Rotate(watermark, self.rotate, tempdir=self.tempdir)
                .use(self._driver)
                .rotate()
            )

        # Set watermark file to be used for upscaling
        try:
            wtrmrk = watermark_file["rotated"]
            watermark_file.update(Info(watermark_file["rotated"]).dimensions)
        except KeyError:
            wtrmrk = watermark_file["path"]

        # 3c. Check if watermark file needs to be upscaled
        scale = 0
        if watermark_file["w"] <= document["w"]:
            scale = float(document["w"] / watermark_file["w"])

            if watermark_file["h"] * scale > document["h"]:
                scale = float(document["h"] / watermark_file["h"])

            watermark_file["upscaled"] = (
                Upscale(wtrmrk, scale=scale, tempdir=self.tempdir)
                .use(self._driver)
                .upscale()
            )
        self.scale = scale
        return watermark_file

    @property
    def _set_filenames(self) -> Tuple[str, str]:
        # 4a. If upscaled PDF file does not exist use input PDF path
        try:
            pdf = self.document["upscaled"]
        except KeyError:
            pdf = self.document["path"]

        # 4b. If rotated watermark file does not exist use input watermark path
        try:
            try:
                watermark = self.watermark_file["upscaled"]
            except KeyError:
                watermark = self.watermark_file["rotated"]
        except KeyError:
            watermark = self.watermark_file["path"]
        return pdf, watermark

    def add(self) -> str:
        """Add watermark to PDF by merging original PDF and watermark file."""
        return self.execute()

    def pypdf(self) -> str:
        watermark_page = PypdfReader(self.wtrmrk_fname).get_page(0)

        writer = PypdfWriter()

        reader = PypdfReader(self.pdf_fname)
        writer.append(reader)

        # width = float(watermark_page.mediabox[2]) / 2
        # height = float(watermark_page.mediabox[3]) / 2

        for content_page in writer.pages:
            # content_page.merge_translated_page(watermark_page, tx=width, ty=height, over=False)
            content_page.merge_page(watermark_page, over=False)

        with open(self.output_filename, "wb") as fp:
            writer.write(fp)

        return self.output_filename

    def pdfrw(self) -> str:
        """Faster than PyPDF3 method by as much as 15x."""
        # TODO: Fix issue where watermark is improperly placed on large pagesize PDFs
        # print(Info(document).size)
        # print(Info(watermark).size)
        # print('\n')

        # Open both the source files
        wmark_trailer = PdfReader(self.wtrmrk_fname)
        trailer = PdfReader(self.pdf_fname)

        # Handle different sized pages in same document with
        # a memoization cache, so we don't create more watermark
        # objects than we need to (typically only one per document).

        wmark_page = wmark_trailer.pages[0]
        wmark_cache = {}

        # Process every page
        for pagenum, page in enumerate(trailer.pages, 1):

            # Get the media box of the page, and see
            # if we have a matching watermark in the cache
            mbox = tuple(float(x) for x in page.MediaBox)
            odd = pagenum & 1
            key = mbox, odd
            wmark = wmark_cache.get(key)
            if wmark is None:

                # Create and cache a new watermark object.
                wmark = wmark_cache[key] = PageMerge().add(wmark_page)[0]

                # The math is more complete than it probably needs to be,
                # because the origin of all pages is almost always (0, 0).
                # Nonetheless, we illustrate all the values and their names.

                page_x, page_y, page_x1, page_y1 = mbox
                page_w = page_x1 - page_x
                page_h = page_y1 - page_y  # For illustration, not used

                # Scale the watermark if it is too wide for the page
                # (Could do the same for height instead if needed)
                if wmark.w > page_w:
                    wmark.scale(1.0 * page_w / wmark.w)

                # Always put watermark at the top of the page
                # (but see horizontal positioning for other ideas)
                wmark.y += page_y1 - wmark.h

                # For odd pages, put it at the left of the page,
                # and for even pages, put it on the right of the page.
                if odd:
                    wmark.x = page_x
                else:
                    wmark.x += page_x1 - wmark.w

                # Optimize the case where the watermark is same width
                # as page.
                if page_w == wmark.w:
                    wmark_cache[mbox, not odd] = wmark

            # Add the watermark to the page
            PageMerge(page).add(wmark, prepend=self.underneath).render()

        # Write out the destination file
        PdfWriter(self.output_filename, trailer=trailer).write()
        return self.output_filename



import os
from typing import List

from PyBundle import bundle_dir, resource_path


def _image_directory() -> str:
    directory = os.path.join(bundle_dir(), "img")
    if os.path.exists(directory):
        return directory
    else:
        raise Exception("No image directory found")


IMAGE_DIRECTORY = _image_directory()


def available_images() -> List[str]:
    images = [i for i in os.listdir(IMAGE_DIRECTORY) if not i.startswith(".")]
    if len(images) > 0:
        return sorted(images, reverse=True)
    else:
        return []


IMAGE_DEFAULT = resource_path("watermark.png")

__all__ = ["IMAGE_DEFAULT", "IMAGE_DIRECTORY", "available_images"]
